#coding: iso-8859-1;
############################################################################
#                                                                          #
#    This file 'SPyRO.py'                                                  #
#    is part of 'SPyRO: Simple Python Remote Objects'                      #
#    Copyright (C) 2004-2005 by Eric Sadit Tellez Avila                    #
#    Copyright (C) 2005-2006 by Eric Sadit Tellez Avila                    #
#    Copyright (C) 2007 by Eric Sadit Tellez Avila                         #
#    Copyright (C) 2008 by Eric Sadit Tellez Avila                         #
#    sadit@lsc.fie.umich.mx or donsadit@gmail.com                          #
#                                                                          #
#    is part of 'Dweba: Distributed Web Application Framework (for Python)'#
#    Copyright (C) 2008      by Eric Sadit Tellez Avila                    #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################

"""
Simple Python Remote Objects
Implements Simple Remote Objects (attributes and methods).

Provides the complete suite: server and client services, and RemoteObject
wrappers to objects.

Implement new formats are quite simple. Just create a class or module with
the methods or functions loads and dumps and register your interface
with the |registerFormat| function. Thread safe operation is responsibility
of the new add-on.

Important Note:
Since SPyRO |Server| is multithreaded, but |CommunicationCliente| had synchronous
protocol, it will not be able to manage many request (every action over a
object is a request) on the same connection. It is necessary use a |Client|
object peer concurrent thread, in the best case, may be one or two are
necessary or use |TClient| (uses a pool of |Client|'s transparentely).
 
"""

import thread
from threading import Lock
__thread_data__ = {}
__thread_exit_handler__ = {}

def getrequest():
    """ Returns the spyro's request dictionary that
    rules the current thread or spyro request. If this thread is not
    ruled by an spyro request, it will raise a KeyError exception"""
    return __thread_data__[thread.get_ident()]

def _setrequest(data):
    __thread_data__[thread.get_ident()] = data

def _delrequest():
    ident = thread.get_ident()
    try:
        handlers = __thread_exit_handler__[ident]
    except KeyError:
        pass
    else:
        for h in handlers:
            try:
                h()
            except Exception, e:
                warn.warn("Error in SPyRO _delrequest exit handler: ", repr((ident, h, e)))
        del __thread_exit_handler__[ident]
    del __thread_data__[ident]

def register_exit_handler(handler):
    ident = thread.get_ident()
    try:
        __thread_exit_handler__[ident].append(handler)
    except KeyError:
        __thread_exit_handler__[ident] = [handler]

import warn
warn = warn.Warn("SPyRO.SPyRO")
import os
import sys
import socket, SocketServer
import gzip
try:
    from cStringIO  import StringIO
except ImportError:
    from StringIO import StringIO

from SocketServer import ThreadingTCPServer
import Queue
import httplib
import urllib2
import time

import shttp
import shttp.http as http

from shttp import uri
from shttp.uri import URI
import policies
# from policies import AuthError, AuthorizationRequest, AuthPasswordRequest
from policies import AuthError
# AuthPasswordRequest
from codepools import CodePool, ServerObjectPool, ModulePool

try:
    import OpenSSL
    import OpenSSL.SSL as SSL
except ImportError, err:
    warn.debug("Unable to use HTTPS as transport layer ", err)
    SSL = None

import args as Args

sleep_min_sec = 0.05
sleep_max_sec = 100.0

SendByReference = 'REF'
SendByValue = 'VAL'

class RemoteException(Exception): pass
class AttributeIsCallable(Exception): pass

import types
from types import DictType

def _set_booleans():
    BooleanType = types.IntType
    __builtins__['True'] = 1
    __builtins__['False'] = 0

try:
    BooleanType = types.BooleanType
    True
except:
    warn.warn("An old version of python or jython was detected")
    _set_booleans()

_HIBRID_SEND_VALUE_tuple = (types.StringType,
                            types.NoneType,
                            types.IntType,
                            BooleanType, 
                            types.LongType, types.FloatType,
                            types.UnicodeType,
                           #types.ComplexType,
                            )
def defaultoverture(environment):
    """ By default pickle is disabled, to enable you must overload 'overture' or
    set to None
    
    Any other prelude action must be done in this method. The overture is provided
    to control the execution before start it. The authorities (from policies) are
    checked when a protocol is up and loading. This method must be set to control
    access and tune settings before any protocol and authority.
    """
    q = environment['QUERY_STRING'].lower()
    if 'pickle' in q:
        raise Exception("The pickle marshaler is not allowed on this SPyRO server")


import formats
orb_formats = formats.ORBFormats()
orb_formats.init()



def _choose_format(format, default_format):
    return orb_formats.getFormat(format)

class TransportProtocolMixIn:
    """ Common methods in the ProtocolHTTP* server and clients"""
    def _choose_format(self, format, default_format):
        if default_format and not format:
            format = default_format
        if format == self.format:
            return
        self.format = format
        self.broker = orb_formats.getFormat(format)

    def _load_result(self, x):
        # warn.debug('x'*3,' LOAD_RESULT\n',x)
        return Response(**x)

class ProtocolHTTPReader(TransportProtocolMixIn):
    """ Protocol used by clients, it runs under HTTP transmition layer """
    def __init__(self, address, location, compress, usingssl=False):
        """ Initialize a client protocol manager.

        Using |address| as a tuple (host,port) and |location|
        Set the remote object location to perform the remote operations.
        In |location| the format of the broker is attached concatenating
        a '?' and a name of the format to the normal location.
        |compress| Use gzip compression in messages
        |usingssl| If True, the connection will be done using https
        
        For example, to request a remote server that is listening on
        /path/SPyRO to use the registered XMLRPC format /path/SPyRO?XMLRPC"""
        self.address = address
        self.usingssl = usingssl
        if location[0] == '/':
            location=location[1:]
        else:
            raise Exception("Location is not absolute it must start with / %"%repr(location))
        self.location = location
        self.__closed = False
        self.format = None
        self._choose_format(uri.query_string(location), None)
        self.compress = compress

    def get_req(self):
        if self.usingssl:
            proto = "https"
        else:
            proto = "http"
        host = "%s:%s"%(self.address[0],self.address[1])
        url = "%s://%s/%s"%(proto, host, self.location)
        headers = {'Host': host}
        if self.compress:
            headers['Accept-Encoding']='x-gzip'
        # reload(urllib2)
        return urllib2.Request(url, headers = headers)
        
    def __del__(self):
        # if not self.__closed: self.close()
        pass

    def close(self):
        # self.__closed = True
        pass

    def __request(self, msg, contenttype = "application/xml"):
        """ Send the message |msg|"""
        # if self.compress: headers['Accept-Encoding'] = 'x-gzip'
        # self.send_request(self.location, msg, headers=headers)
        req = self.get_req()
        req.add_data(msg)
        req.add_header("Content-Type", contenttype)
        f = urllib2.urlopen(req)
        info = f.info()
        gzipheader = info.getheader("content-encoding")
        data = f.read()
        f.close()
        import md5
        if gzipheader and gzipheader.find("gzip") > 0:
            f = gzip.GzipFile(fileobj=StringIO(data), mode="rb")
            self.msg = f.read()
            f.close()
        else:
            self.msg = data
        
    def get_result(self):
        """ Read a message and returns. It uses a specific format """
        x = self.broker.load_result(self.msg)
        return self._load_result(x)

    def _fixauth(self, msg):
        try:
            msg['auth'] = msg['auth'].__dict__
        except:
            pass
        
    def send_control(self, msg):
        self._fixauth(msg)
        self.__request(self.broker.dump_control(msg))
    
    def send_setAttr(self, msg):
        """ send a setAttr request """
        self._fixauth(msg)
        self.__request(self.broker.dump_setAttr(msg))

    def send_getAttr(self, msg):
        """ send a getAttr request """
        self._fixauth(msg)
        self.__request(self.broker.dump_getAttr(msg))

    def send_callMethod(self, msg):
        """ send a callMethod request """
        self._fixauth(msg)
        self.__request(self.broker.dump_callMethod(msg))


##############################################################################
##############################################################################
#####  The request class
##############################################################################
##############################################################################

class Request:
    """ Object to request to the remote object server

    Values of the variables:
    # Say bye
    ReqBye = 'BYE'
    # Request to Get Remote Attr
    ReqGetAttr = 'GET'
    # Set an attribute
    ReqSetAttr = 'SET'
    # Del an object
    ReqDelObj = 'DEL'
    # Request to Call a remote method
    ReqCallMethod = 'CALL'
    """
    # Say bye
    ReqBye = 'BYE'
    # Request to Get Remote Attr
    ReqGetAttr = 'GET'
    # Set an attribute
    ReqSetAttr = 'SET'
    # Del an object
    ReqDelObj = 'DEL'
    # Request to Call a remote method
    ReqCallMethod = 'CALL'

def newRequest(objname, attrname, reqtype, auth):
    return {'objname' : objname, 'attrname' : attrname, 'reqtype' : reqtype, 'auth' : auth}

def newRequestSet(objname, attrname, value, local_server, auth):
    """ Creates a new Request for a |setattr| operation, to set the
    attrname to value"""
    req = newRequest(objname, attrname, Request.ReqSetAttr, auth)
    req['value'] = value
    req['local_server'] = local_server
    return req

def newRequestGet(objname, attrname, rettype, auth):
    """ Creates a new Request for a |getattr| operation """
    req = newRequest(objname, attrname, Request.ReqGetAttr, auth)
    req['rettype'] = rettype
    return req

def newRequestDel(objname, auth):
    """ Creates a new Request for a DEL object operation """
    return newRequest(objname, None, Request.ReqDelObj, auth)

def newRequestBye():
    """ Creates a new Request for a BYE object operation """
    return newRequest(None, None, Request.ReqBye, None)

def newRequestCall(objname, attrname, args, kwargs, rettype, local_server,
                   auth, extra):
    """ A SPyRO request, call mode
    |objname| Object name
    |attrname| The methodname
    |args| Positional arguments, tuple
    |kwargs| Keyword arguments, dictionary
    |rettype| The return type, SendByReference or SendByValue
    |local_server| If Request to call a remote method and exists port we call
    the method with the pass the arguments by reference the server receives the
    name or id in the remote peer (the caller) and a port to connect. The IP
    address is obtained by the socket
    |auth| The authentication object
    |extra| Additional info to append to message, for example routed calls
    """
    req = newRequest(objname, attrname, Request.ReqCallMethod, auth)
    req['args'] = args
    req['kwargs'] = kwargs
    req['rettype'] = rettype
    req['local_server'] = local_server
    if extra:
        req['extra'] = extra
    return req

##############################################################################
##############################################################################
#####  The response class set
##############################################################################
##############################################################################

def ObjectWrapper_GuessType(xobject, xserver):
    """ Function to wrap object guessing the more convenient type to send
    (send by reference or by value) """
    if type(xobject) in _HIBRID_SEND_VALUE_tuple:
        return ObjectWrapper_ByValue(xobject)
    else:
        return ObjectWrapper_ByReference(xobject, xserver)

def ObjectWrapper_ByReference(xobject, xserver, xauthority = None):
    """ Wrap objects to send objects to Reference """
    xobjname = '!' + str(id(xobject))
    xserver.setsend(xobjname, xobject, xauthority)
    return (xobjname, SendByReference)

def ObjectWrapper_ByValue(xobject):
    """ Wrap objects to send objects to Reference """
    return (xobject, SendByValue)

def ObjectWrapper_Unpack(objarg):
    return objarg

def ObjectWrapper_GetObject(objarg):
    return objarg[0]

def ObjectWrapper_GetType(objarg):
    return objarg[1]

def newResponse(errcode, retvalue, rettype):
    """ Create and returns a server side response from the arguments """
    return {'errcode': errcode, 'retvalue' : retvalue, 'rettype': rettype}


class Response:
    """ Objects to send as response at requests.
    The request codes are the following:
    Successful = 0
    # Unknown Object, no object registered to share in server, retvalue is None
    ErrorUnknownObject = 1
    # Object doesn't have specified attribute, retvalue = None
    ErrorUnknownAttribute = 2 # NO USAR
    # An Exception Throw, exception is returned in retvalue
    ErrorExceptionThrowed = 3
    # An error on protocol
    ErrorUnknownRequestType = 4
    # The object is callable, we don't wanna send it, we want to call it
    ObjectIsCallable = 5
    # The object or the method is not accesible with the authentication or is not valid
    ErrorInTheAuthentication = 6
    """
    # No error, successful call
    Successful = 0
    # Unknown Object, no object registered to share in server, retvalue is None
    ErrorUnknownObject = 1
    # Object doesn't have specified attribute, retvalue = None
    ErrorUnknownAttribute = 2 # NO USAR
    # An Exception Throw, exception is returned in retvalue
    ErrorExceptionThrowed = 3
    # An error on protocol
    ErrorUnknownRequestType = 4
    # The object is callable, we don't wanna send it, we want to call it
    ObjectIsCallable = 5
    # The object or the method is not accesible with the authentication or is not valid
    ErrorInTheAuthentication = 6

    ErrorNames = ['Successful', 'ErrorUnknownObject','ErrorUnknownAttribute', 
    'ErrorExceptionThrowed', 'ErrorUnknownRequestType','ObjectIsCallable','ErrorInTheAuthentication']

    def __init__(self, errcode = None, retvalue = None, rettype = None):
        """ Initialize the response with an error code  |errcode| and the
        return value |retvalue|"""
        self.errcode = errcode
        self.retvalue = retvalue
        self.rettype = rettype

    def errname(self):
        """ Returns the name of the errcode """
        return Response.ErrorNames[self.errcode]
    
    def prepareToCall(self, remoteobj):
        """ Returns the returned value of calling |methodname| with arguments
        |args|. Can raise remote exceptions or locals due to remote errors """
        #self.attrname = self.retvalue.attrname
        #print "**********> PrepareToCall: ",self.retvalue
        self.attrname = self.retvalue[1] #'attrname'
        #self.remoteobj = remoteobj
        self.comm,self.objname,self.objauth = remoteobj.__xxxconnxxx__
        # preparing cache of callable objects
        remoteobj.__dict__[self.attrname] = self.__call__

    def __call__(self, *args, **kwargs):
        """ Internal function, this is the real __call__ method """
        return self.comm.callMethodGuess(self.objname, self.attrname,
                                         args, kwargs, SendByValue,
                                         self.comm.local_server,
                                         auth = self.objauth)

class BasicServer:
    """ The base of spyro's communication servers """
    def __init__(self,
                 serverurl = "http://localhost:8000/spyro/spyro.py?MARSHAL",
                 priorities = None,
                 objectpool = None,
                 overture = defaultoverture):
        """ |priorities| Priorities of the protocols
            |objectpool| The object pool manager, if it's None then a new SeverObjectPool
                 instance is used
        """
        self.serverurl = serverurl
        self.priorities = priorities
        # Open connections
        # self.connections = {} 
        self.commclients = {}
        # Number of registered returns
        self.lockconn = Lock()
        if not objectpool:
            objectpool = ServerObjectPool()
        self.setobj = objectpool.setobj
        self.getobj = objectpool.getobj
        self.delobj = objectpool.delobj
        self.setsend = objectpool.setsend

    def __getstate__(self):
        """ To pickle functions """
        return {}

    def regconn(self, conn):
        """ Register a client connection """
#         self.lockconn.acquire()
#         self.connections[conn] = 1
#         self.lockconn.release()
        pass

    def delconn(self, conn):
        """ Removes a client connection """
        pass
#         self.lockconn.acquire()
#         del self.connections[conn]
#         self.lockconn.release()

    def __setitem__(self,k,v):
        """Register an object. """
        self.setobj(k,v)

    def set(self,k,v):
         """ Register an object |v| with the ID |k|. Use |setobj| instead"""
         warn.stack("This method is deprecated")
         raise "This method is deprecated"
    
    def close(self):
        """ Close server """
        pass

    ######################################################################
    ### Object's execution block
    ######################################################################
    def getcomm(self, local_server):
        """ Returns the CommunicationClient associated to the address that
        made the |request| through the |serv| """
        # addr=(xxxself.request.getpeername()[0],l[0])
        warn.warn("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX", localserver)
        assert False, "XXXXXXXXXXXXXXXXXXXX"
        addr=(local_server[2],local_server[0])
        try:
            comm = self.commclients[addr]
        except KeyError:
            comm = CommunicationClientThreaded(addr, local_server[1], local_server=self, poolSize=5)
            self.commclients[addr] = comm
        return comm

    #def __del__(self):
    #    warn.log("DEL PEERCONNECTION", self.__dict__)
    def getargument(self, comm, arg):
        """ Get the argument asociated to |arg| using the |comm|
        It manages references and values """
        argobj, argtype = ObjectWrapper_Unpack(arg)
        if argtype == SendByReference:
            return comm[argobj]
        else:
            return argobj

    def handlecall(self, obj, atr, objname, methodname, args, kwargs, local_server):
        """ handles the execution of methods. called by handle.
        If the object has the attribute __spyro__ it must be an object
        with the |__contains__| method and every element is a method
        that must be handled by pass (call methods with the request
        method as the first argument, other arguments are passed as normal)

        If the object has the attribute '__spyro_formats__' it must be a
        dictionary (object with the __getitem__ method) of { methodname: argumentformats }

        Where |methodname| must be the name of methods where the arguments need to be checked
        |argumentformats| is the format as listed in the SPyRO.args module
        """
        # methodname = request.attrname
        try:
            hasspyro = methodname in getattr(obj,'__spyro__')
        except AttributeError:
            hasspyro = False
        try:
            fmt_args = getattr(obj,'__spyro_formats__')
            fmtcomp = None
            try:
                comp = fmt_args['!compiled!']
                # print repr(comp),"<","l" * 70, ">"
                try:
                    fmtcomp = comp[methodname]
                except (KeyError, AttributeError):
                    fmtcomp = None
            except KeyError:
                fmt_args['!compiled!'] = {}
            if fmtcomp is None:
                try:
                    fmtcomp = Args.Arguments(fmt_args[methodname])
                    fmt_args['!compiled!'][methodname] = fmtcomp
                except KeyError:
                    fmtcomp = None
            if not(fmtcomp is None):
                fmtcomp.check(args, kwargs)
        except AttributeError:
            pass
        if hasspyro:
            warn.debug("A new api to __spyro__ was developed... check your implementation please")
            return atr(self, objname, attrname, args, kwargs)
        if local_server:
            comm = self.getcomm(local_server)
            args = list(args)
            for index, arg in enumerate(args):
                args[index] = self.getargument(comm, arg)
            # args = tuple(_args)
            for k,v in kwargs.items(): # keyword arguments
                kwargs[k] = self.getargument(comm, v)
        if not (type(kwargs) is DictType): kwargs = {}
        return atr(*args, **kwargs)

    def can_del(self, authobj, req):
        if authobj == None: return True
        if authobj.can_del(req): return True
        raise AuthError, "Can't del object %(objname)s permission denied"%req

    def can_set(self, authobj, req):
        if authobj == None: return True
        if authobj.can_set(req): return True
        raise AuthError, "Can't set attr '%(attrname)s' in the object '%(objname)s'"%req

    def can_follow_objname(self,  authobj, req):
        if authobj == None: return True
        if authobj.can_follow_objname(req): return True
        raise AuthError, "Can't follow using dot names in the object '%(objname)s'"%req

    def can_get(self, authobj, req):
        if authobj == None: return True
        if authobj.can_get(req): return True
        raise AuthError, "Can't get attr '%(attrname)s' in the object '%(objname)s'"%req

    def can_call(self, authobj, req):
        if authobj == None: return True
        if authobj.can_call(req): return True
        raise AuthError, "Can't call in the object '%(objname)s'"%req

    def can_register(self, authobj, req):
        if authobj == None: return True
        if authobj.can_register(req): return True
        raise AuthError, "Can't register objects, permission denied"

    def handle_http(self, input_data, format, environment, remote_server=None, overture=None, request_rewrite=None):
        """ Handles an spyro request reading the request from rfile, using the broker specified in environment['QUERY_STRING']
        It writes data using wfile and the broker.
          |input_data| message
          |format| The format of this request
          |environment| Environment of the request
          |remote_server| In case of arguments passed by reference
          
        Optional arguments:
          |overture| Overture to be executed before anything. If overture throws an exception
          solve_request will not be executed. Prototype
               overture(environment)
          |request_rewrite| If it's different of None it will be executed after the read of the
          request and the generated dictionary will be sent and perhaps modified.
             Prototype
             request_rewrite(requestdict,basicserver,environment)
        """
        if overture: overture(environment)
        broker = orb_formats.getFormat(format)
        self.regconn(self)
        req = broker.load_request(input_data)
        try:
            local_server = req['local_server']
            if local_server is None: raise KeyError
            req['local_server'] = remote_server
        except KeyError:
            pass
        if request_rewrite:
            request_rewrite(req, self, environment)
        res = self.handle_request(**req)
        msg = broker.dump_result(res)
        # print ">>>>>>>>>>>>>>>>>>>>>>> MSG:\n", msg
        return msg

    def handle_request(self, *args, **kwargs):
        """ Handles one spyro's request. Independent of transport layer.
        See _handle_request for argument description
        """
        try:
            res = self._handle_request(*args, **kwargs)
        except AuthError, error:
            res = newResponse(Response.ErrorInTheAuthentication,str(error),SendByValue)
        except (Exception, types.StringType, types.UnicodeType, types.NoneType, BooleanType,
                types.LongType, types.FloatType, types.TupleType, types.ListType, types.DictType):
            etype, evalue, etraceback = sys.exc_info()
            error = "%s %s" % (etype, evalue)
            warn.warn(error)
            res = newResponse(Response.ErrorExceptionThrowed,str(error),SendByValue)
        except:
            _delrequest()
            raise
        _delrequest()
        return res
        
    def _handle_request(self, objname, attrname, args = (), kwargs = {}, auth = None, extra = None, local_server = None,
                       reqtype=Request.ReqCallMethod, rettype=SendByValue, value=None, req=None):
        if req is None: req = {'objname' : objname, 'attrname' : attrname, 'args' : args, 'kwargs' : kwargs,
                               'auth' : auth, 'local_server' : local_server, 'reqtype' : reqtype, 'rettype' : rettype,
                               'value' : value, 'req' : None
                               }
        _setrequest(req)
        ##########################
        ### Say bye
        ##########################
        if Request.ReqBye == reqtype:
            return newResponse(Response.Successful, None, SendByValue)
        ##########################
        ### Object name resolution
        ##########################
        try:
            obj, authobj = self.getobj(objname)
        except KeyError:
            try:
                # Resolve object name: attribute of objects
                objpath = objname.split('.')
                obj, authobj = self.getobj(objpath.pop(0))
                if not self.can_follow_objname(authobj, req):
                    raise AttributeError, "UnknownObject because: Forbidden access to object's attribute"
                for oname in objpath:
                    obj = getattr(obj, oname) # authobj is inherited
            except AttributeError:
                return newResponse(Response.ErrorUnknownObject, 'Unknown Object: %s'%repr(objname), SendByValue)
        req['obj'] = obj
        #############################################
        ### Deleting objects (generated by users)
        #############################################
        if Request.ReqDelObj == reqtype and self.can_del(authobj, req):
            if objname == '!':
                warn.debug("-DEL-OBJ-", objname)
                #serv.delobj(objname)
            return newResponse(Response.Successful, '', SendByValue)
        #############################################
        ### The 'set' action
        #############################################
        if Request.ReqSetAttr == reqtype and self.can_set(authobj, req):
            if local_server != None:
                value=self.getargument(self.getcomm(local_server), value)
            setattr(obj, attrname, value)
            return newResponse(Response.Successful, None, SendByValue)
        val=None
        #############################################
        ### The 'get' action
        #############################################
        atr=getattr(obj, attrname)
        if Request.ReqGetAttr == reqtype and self.can_get(authobj, req):
            if callable(atr):
                return newResponse(Response.ObjectIsCallable,(objname,attrname), SendByValue)
            #print "x"*30,">", repr(atr)
            val = atr
        #############################################
        ### The 'call' action
        #############################################
        elif Request.ReqCallMethod == reqtype and self.can_call(authobj, req):
            val = self.handlecall(obj, atr, objname, attrname, args, kwargs, local_server)
        else:
            return newResponse(Response.ErrorUnknownRequestType, 'Unknown Request Type: %s'%repr(reqtype))
        if rettype == SendByReference and self.can_register(authobj, req):
            res = newResponse(Response.Successful, self.setsend("?%s"%id(val),val), SendByReference)
        else:
            res = newResponse(Response.Successful, val, SendByValue)
        return res

Server = BasicServer

##############################################################################
##############################################################################
#####  The client class set
##############################################################################
##############################################################################

def default_socket_error_handler(comm, _socket, addr, error):
    """ The default, fool, error handler in Client connections """
    errno, errstr = error
    if errno == 111:
        seconds = sleep_min_sec
        connected = False
        while True:
            try:
                _socket.connect(addr)
                break
            except socket.error, _error:
                if _error[0] != 111:
                    default_socket_error_handler(comm, _socket, addr, _error)
                    break
            # WISH: Try to find another server or port and connect,
            # sleep if we can't find more servers
            warn.debug("SPyRO Client to %s: %s"%(addr, error),
                       "trying in %f seconds"%(seconds))
            time.sleep(seconds)
            if seconds > sleep_max_sec:
                seconds = sleep_min_sec
            else:
                seconds = seconds + seconds
    else:
        warn.warn(error)
        raise error
    warn.debug("Done")

class CommClientMixIn:
    """ The mixin to CommunicationClient{,Threaded}"""
    def get(self, name, auth = None):
        """ Get a local representation of the remote object with ID |name|. Deprecated. Use |getobj| instead"""
        warn.stack("This is not an error, but this method is deprecated")
        return RemoteObject(self, name, auth)

    def __getitem__(self,name):
        """ Get a local representation of the remote object with ID |name|."""
        return RemoteObject(self, name, None)

    def getobj(self, name, auth = None):
        """ Get a local representation of the remote object with ID |name|"""
        return RemoteObject(self, name, auth)

    def getconn(self):
        """ Get a communication object """
        return self
        #raise "Don't use this method (getconn) directly"

    def putconn(self, conn):
        """ Return a communication Object """
        pass
        #raise "Don't use this method (putconn) directly"

    def handleRequest(self, conn, request, send_method):
        """ Common code to getAttr, setAttr and callMethod """
        #conn = self.getconn()
        try:
            send_method(request)
            response = conn.broker.get_result()
        except socket.error:
            warn.debug("======= Unable to connect remote peer in SPyRO.handleRequest")
            conn._do_bye = False
            self.putconn(conn)
            conn.close()
            raise
        except:
            warn.warn("======== Unexpected exception in SPyRO.CommunicationClient*")
            conn._do_bye = False
            conn.close()
            self.putconn(conn)
            raise
        self.putconn(conn)
        if response.errcode == Response.Successful:
            if response.rettype == SendByReference:
                response.retvalue = self[response.retvalue]
            return response
        if response.errcode == Response.ObjectIsCallable:
            raise AttributeIsCallable, response
        if response.errcode == Response.ErrorExceptionThrowed:
            #warn.stack("RemoteException", response.retvalue)
            #warn.stack("RemoteException", type(response))
            raise RemoteException, response.retvalue
        if response.errcode == Response.ErrorInTheAuthentication:
            raise AuthError(response.retvalue)
        raise RuntimeError, "Remote send code %d (%s)\n%s\n"%(response.errcode,
                            response.errname(), request)

    def getAttr(self, objname, attrname, rettype = SendByValue,
                auth = None):
        """ Get an attribute |attrname| of the object |objname| with the method
        to send the object |rettype|, the default is SendByValue
        |auth| The authorizing information to access the object
        """
        request=newRequestGet(objname, attrname, rettype, auth)
        conn = self.getconn()
        return self.handleRequest(conn, request, conn.broker.send_getAttr).retvalue

    def setAttr(self, objname, attrname, value, sendtype=SendByValue,
                local_server = None, auth = None):
        """ Set the attribute |attrname| to |value| in the object |objname|.
        The mode to send the value is |sendtype|.
        The |local_server| is used to send by reference
        |auth| The authorizing information to access the object
        """
        if local_server:
            if sendtype == SendByValue:
                value = ObjectWrapper_ByValue(value, sendtype)
            elif sendtype == SendByReference:
                value = ObjectWrapper_ByReference(value, local_server)
            else:
                value = ObjectWrapper_GuessType(value, local_server)
            local_server = self.get_request_local_server(local_server)
        request = newRequestSet(objname, attrname, value, local_server,
                                auth)
        conn = self.getconn()
        return self.handleRequest(conn, request, conn.broker.send_setAttr).retvalue
    
    def byeRequest(self):
        """ Send a bye request (deprecated, now it's ignored) """
        # conn = self.getconn()
        # self.handleRequest(conn, newRequestBye(), conn.broker.send_control)
        return 

    def delObject(self, objname, auth = None):
        """ Get an attribute |attrname| of the object |objname| with the method
        to send the object |rettype|, the default is SendByValue.
        |auth| The authorizing information to access the object
        """
        request=newRequestDel(objname, auth)
        conn = self.getconn()
        return self.handleRequest(conn, request, conn.broker.send_control).retvalue

    def callMethod(self, objname, attrname, args = (), kwargs = {},
                   rettype = SendByValue, local_server = None, auth = None, extra = None):
        """ Call a method. Every argument must be a ObjectWrapper
        with all the necessary information.
        |objname| The object's name that has the method
        |attrname| The name of the object
        |args| The positional arguments of the call
        |kwargs| The keyword arguments of the call
        |rettype| The return type of the result
        |local_server| the local server to send arguments by reference
        |auth| The authorizing information to access the object
        |extra| Extra general information to be passed in the message
        """
        request = newRequestCall(objname, attrname, args, kwargs,
                                 rettype, local_server, auth, extra)
        conn = self.getconn()
        return self.handleRequest(conn, request, conn.broker.send_callMethod).retvalue

    def get_request_local_server(self, local_server):
        """ Return a local_server tuple to make the request """
        return (local_server.server_address[1],local_server.location)
    
    def callMethodGuess(self, objname, attrname, args = (), kwargs = {},
                        rettype = SendByValue, local_server = None,
                        auth = None, extra = None):
        """ Call a method, Guessing the best type to send arguments
        |objname| The object's name that has the method
        |attrname| The name of the object
        |args| The positional arguments of the call
        |kwargs| The keyword arguments of the call
        |rettype| The return type of the result
        |local_server| the local server to send arguments by reference
        |auth| The authorizing information to access the object
        |extra| Additional information to be added to the message
        """
        if local_server:
            warn.debug("Objects references as arguments, GC problem")
            xargs = []
            for arg in args:
                xargs.append(ObjectWrapper_GuessType(arg, local_server))
            args = tuple(xargs)
            for k,v in kwargs.items():
                kwargs[k] = ObjectWrapper_GuessType(v, local_server)
            local_server=self.get_request_local_server(local_server)
        return self.callMethod(objname, attrname, args, kwargs, rettype,
                            local_server, auth = auth, extra = extra)
    

class CommunicationClientBase(CommClientMixIn):
    """ A base class to any CommunicationServer client. Represents a
    communication channel
    with the server """
    def __init__(self, serv_addr, path, local_server=None,
                 socket_error_handler = default_socket_error_handler,
                 compress = False, usingssl = False):
        """ Initialize client.
        |serv_addr| The remote server address (host,port)
        |path| Location of the server (to apache or another general purpose
               http servers). It describes the encoder protocol using the
               query_string '?' notation.
        |local_server| If is set to a CommunicationServer, when a method is
               called the arguments are passed by reference and the remote
               server
               will try to connect to this server (port and location).
               The IP address is the used to contact the remote server.
        |socket_error_handler| The error handler, called when an exception
            is arised in the connection, parsing, or interpreting messages.
        |compress| Compress the messages between peers
        |usingssl| If True, every connection is performed with the https protocol
        """
        self.socket_error_handler = socket_error_handler
        self.path = path
        self.local_server = local_server
        self.__closed = False
        self.broker = ProtocolHTTPReader(serv_addr, path, compress, usingssl = usingssl)
        self._do_bye = True
        
    def __getstate__(self):
        """ To pickle functions """
        return {'path': self.path }
    
    def close(self):
        """ Close the client """
        self.__closed = True
        if self._do_bye:
            self.byeRequest()
        self.broker.close()

    def __del__(self):
        """ called for the garbage collection system. Close the object
        if it is not closed yet.  """
        if not self.__closed:
            self.close()

class CommunicationClient(CommunicationClientBase):
    """ A Client to CommunicationServer. Represents a communication channel
    with the server """
    def __init__(self, serv_addr, path, local_server=None,
                 socket_error_handler = default_socket_error_handler,
                 compress = False,
                 usingssl = False):
        CommunicationClientBase.__init__(self, serv_addr, path, local_server,
                                         socket_error_handler,
                                         compress = compress,
                                         usingssl = usingssl)
        self.connlock = Lock()
        
    def __getstate__(self):
        """ To pickle functions """
        return CommunicationClientBase.__getstate__(self)

    def getconn(self):
        self.connlock.acquire()
        return self

    def putconn(self, conn):
        self.connlock.release()
    
Client = CommunicationClient

class CommunicationClientThreaded(CommClientMixIn):
    """ The thread safe CommunicationClient. It uses several
    CommunicationClient to manage threading """
    def __init__(self, serv_addr, path, local_server = None, poolSize=0,
                 socket_error_handler = default_socket_error_handler,
                 compress = False,
                 usingssl = False):
        """ Create the threaded client.
        |serv_addr| The server address (host,port).
        |path| The path where is listening the remote host, it has the
        protocol using query_string '?' notation.
        |poolSize| The maximum size of the pool of CommunicationClient objects
        |local_server| If is set to a CommunicationServer, when a method is
        called the arguments are passed by reference and the remote server
        will try to connect to this server (port and location).
        The IP address is the used to contact the remote server.
        |socket_error_handler| The socket error handler.
        |compress| If True it uses gzip compression to content
        |usingssl| If True it uses the https protocol to perform connections
        """
        self.queue = Queue.Queue(poolSize)
        self.path = path
        self.local_server = local_server
        self.ehandler = socket_error_handler
        self.address = serv_addr
        self.__closed = False
        self.compress = compress
        self.usingssl = usingssl
    
    def __getstate__(self):
        """ To pickle functions """
        return {'path': self.path, 'address': self.address}
    
    def getconn(self):
        try:
            comm=self.queue.get_nowait()
        except Queue.Empty:
            comm=CommunicationClientBase(self.address, self.path,
                                         self.local_server,
                                         socket_error_handler = self.ehandler,
                                         compress = self.compress,
                                         usingssl = self.usingssl)
        return comm

    def putconn(self, conn):
        try:
            self.queue.put(conn)
        except Queue.Full:
            pass

    def __del__(self):
        if not self.__closed:
            self.close()
            
    def close(self):
        while not self.__closed:
            try:
                comm = self.queue.get_nowait()
                comm.close()
            except Queue.Empty:
                break
        self.__closed = True

TClient = CommunicationClientThreaded

class CallClient(Client):
    """ Special Client thay always call methods, no attribute operations are allowed"""
    def getAttr(self,objname, attrname, rettype = SendByValue, auth = None):
        raise AttributeIsCallable, Response(Response.ObjectIsCallable, retvalue = (objname, attrname), rettype = rettype)


class TCallClient(TClient):
    """ Special TClient to always call methods, no attribute operations"""
    def getAttr(self,objname, attrname, rettype = SendByValue, auth = None):
        raise AttributeIsCallable, Response(Response.ObjectIsCallable, retvalue = (objname, attrname), rettype = retttype)

##############################################################################
##############################################################################
##### Remote Object Class and Helper methods
##############################################################################
##############################################################################

class FakeIterator:
    """ FakeIterator """
    def __init__(self, obj):
        self.obj = obj
        self.pos = -1
        self.len = len(obj) - 1

    def next(self):
        if self.pos < self.len:
            self.pos = self.pos + 1
            return self.obj[self.pos]
        else:
            raise StopIteration

    def __call__(self):
        return self


class RemoteObject:
    """ Local representation of a RemoteObject. The remote object needs to be
    registered in the server. """
    def __init__(self, connection, objectname, auth):
        self.__dict__['__xxxconnxxx__'] = (connection, objectname, auth)

    def __getstate__(self):
        return {}
    
    def __getattr__(self, attrname):
        """ Returns the remote attribute named |attribute| as local object.
        Can raise remote exceptions or locals due remote errors """
        if attrname == '__iter__':
            #return lambda: (self[i] for i in range(0,len(self)))
            return FakeIterator(self)
        conn,objname,auth = self.__xxxconnxxx__
        try:
            return conn.getAttr(objname, attrname, SendByValue,
                                auth = auth)
        except AttributeIsCallable, response:
            r = response.args[0]
            r.prepareToCall(self)
            return r

    def __del__(self):
        conn, objname, auth = self.__xxxconnxxx__
        if objname[0] == '!':
            warn.debug("-USR-DEL-OBJ-",objname)
            conn.delObject(objname, auth)
    
    def __setattr__(self, attrname, value):
        """ __setattr__('name', value) <==> x.name = value. Where |x| is
        a remote object """
        conn,objname, auth = self.__xxxconnxxx__
        return conn.setAttr(objname, attrname, value, SendByValue, None,
                            auth = auth)

    def __repr__(self):
        """ Internal representation """
        return "<SPyRO.RemoteObject name: %s, local id: %s>"%(getobjname(self),
                                                              hex(id(self)))

#     def __eq__(self, arg):
#         comm, objname, auth = self.__xxxconnxxx__
#         return comm.callMethodGuess(objname,  '__eq__', (arg), {},
#                                     SendByValue, None, auth = auth)
        
#     def __cmp__(self, arg):
#         comm, objname, auth = self.__xxxconnxxx__
#         return comm.callMethodGuess(objname, '__cmp__', (arg), {},
#                                     SendByValue, None, auth = auth)

def getobject(address, location, name, local_server = None, poolSize=0, auth=None):
    """ Can be used as a shortcut when we need just one object by connection """
    comm=CommunicationClientThreaded(address, location, local_server, poolSize)
    return comm.get(name, auth)

def getconn(remoteobject):
    """ Return the connection of the remote object"""
    return remoteobject.__xxxconnxxx__[0]

def getobjname(remoteobject):
    """ Returns the name of the object """
    return remoteobject.__xxxconnxxx__[1]
