#coding: iso-8859-1;
############################################################################
#                                                                          #
#    This file 'codepools.py'                                              #
#    is part of 'SPyRO: Simple Python Remote Objects'                      #
#    Copyright (C) 2004-2005 by Eric Sadit Tellez Avila                    #
#    Copyright (C) 2005-2006 by Eric Sadit Tellez Avila                    #
#    Copyright (C) 2007 by Eric Sadit Tellez Avila                         #
#    Copyright (C) 2008 by Eric Sadit Tellez Avila                         #
#    sadit@lsc.fie.umich.mx or donsadit@gmail.com                          #
#                                                                          #
#    is part of 'Dweba: Distributed Web Application Framework (for Python)'#
#    Copyright (C) 2008      by Eric Sadit Tellez Avila                    #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################

"""
Code pool for SPyRO

Code pools are set of objects, represented by real objects, files, zip files, etc.

"""
import os
import sys
from threading import Lock
from exec_code import exec_code
import policies
import warn

import types
def _set_booleans():
    BooleanType = types.IntType
    __builtins__['True'] = 1
    __builtins__['False'] = 0

try:
    BooleanType = types.BooleanType
    True
except:
    warn.warn("An old version of python or jython was detected")
    _set_booleans()

##############################################################################
##############################################################################
#####  The server class set
##############################################################################
##############################################################################

class ServerObjectPool:
    """
    The interface to pool of objects. This class is a basic
    pool implemented with a dictionary. To create a different
    pool of objects is necessary to create a class that implement
    every method in this interface.
    """
    def __init__(self, objects = None):
        """
        Constructor
        
        @param objects: Initial set of objects
        @type objects: dict, None
        """
        if objects is None:
            objects = {}
        self.objects = objects

    def setobj(self, name, obj, authobj = None):
        """
        Register an object into the pool
        @param obj: The object to store
        @param name: Name of the object
        @param authobj: The authentication object, see L{SPyRO.policies} or
                L{auth3k.authweb} for a sophisticated extension to L{policies}.
        """
        self.objects[name] = (obj,authobj)

    def getobj(self, name):
        """
        Get an object and its authorization data
        
        @param name: The name of the object to retrieve
        @return: The tuple (object, authobject), or an KeyError exception if the object I{name} don't exists.
        @rtype: tuple (object, authobj)
        """
        return self.objects[name]

    def delobj(self, name):
        """
        Remove an object
        @param name: The name of the object
        @return: Nothing
        """
        del self.objects[name]

    def setsend(self, newid, ret, auth = None):
        """
        Register an object (to send by reference)
        @param newid: The object id (name) to associate to this object
        @param ret: The object to register
        @param auth: The authentication object
        @return: Returns the newid argument
        """
        self.setobj(newid, ret, auth)
        return newid

class ItemCodePool:
    """
    Internal usage. It provides an easy way to construct an object and put
    variables and methods on it.
    """
    pass

class CodeLoader:
    """
    Python code loader abstract class
    """
    def read(self, name):
        """
        Read python code associated with a name
        
        @param name: The name to be readed (it has many means, depends of the codepool)
        @return: Code string
        @rtype: string
        """
        raise KeyError("This method must be overloaded")

    def getmtime(self, name):
        """
        Returns the code's modification time
        @param name: The code's name
        @return: Modification time in seconds from the epoch
        @rtype: float
        """
        return 0

class FileSystemCodeLoader(CodeLoader):
    """
    File system code loader
    """
    def __init__(self):
        """Constructor"""
        pass

    def read(self, filename):
        """Read a piece of code from the filesystem L{CodeLoader}"""
        f = file(filename)
        d = f.read()
        f.close()
        return d

    def getmtime(self, filename):
        """Modification time of a file in the filesystem L{CodeLoader}"""
        return os.stat(filename).st_mtime

class ZipCodeLoader(CodeLoader):
    def __init__(self, zipfile, originseparator='/'):
        """
        Constructor
        @param zipfile: The zipfile object
        @type zipfile: ZipFile L{zipfile}
        @param originseparator: The separator of the filesystem where the zip
                    was created
        """
        self.zipfile = zipfile
        self.originseparator = originseparator

    def read(self, filename):
        """Read a piece of code from a zipfile. L{CodeLoader}"""
        filename = filename.replace(os.path.sep, self.originseparator)
        return self.zipfile.read(filename)

    def getmtime(self, filename):
        """Modification time of a file from a zipfile. L{CodeLoader}"""
        filename = filename.replace(os.path.sep, self.originseparator)
        self.zipfile.getinfo(filename)
        return 0

class CodePool:
    """
    
    Register python modules and executes functions in modules.
    The registration of objects must not be used, instead a list of module's path is given.
    The registrarion is supported to serve setsend calls. The getobj method reads modules
    from path. The module is loaded the first time that is requested, next requests uses
    a cached version of the module. If the module changes in the disk, is reloaded
    (if checkupdates=True, True by default). Every
    time that the module is readed it can execute global statments

    """
    def __init__(self, path, checkupdates=True, codeloader=None):
        """ Init the ServerObjectPool object with a """
        self.path = path
        self.objects = {}
        self.cache = {} # name : timestamp, filename
        self.lock = Lock()
        self.gvars = {}
        if codeloader is None:
            codeloader = FileSystemCodeLoader()
        self.codeloader = codeloader
        self.checkupdates = checkupdates
        
    def setobj(self, name, obj, authobj = None):
        """ Register an object |obj| with the ID |name| """
        self.lock.acquire()
        self.objects[name] = (obj,authobj)
        self.lock.release()

    def _import_(self, name, _globals, _locals, fromlist):
        """
        Replace function of __import__ to support
        import modules from the code pool
        """
        try:
            return self.getobj(name)[0]
            # return imp.load_module(name,*imp.find_module(name,self.path))
        except KeyError,e:
            return __import__(name,_globals,_locals,fromlist)

    def _reload_(self, mod):
        """ Reload an SPyRO module or a python module """
        try:
            import sys
            sys.modules[mod.__name__] = self.getobj(mod.__name__)[0]
        except KeyError:
            reload(mod)
    
    def getobj(self, name):
        """ Get an object and its authorization data, it must raise a
        KeyError exception if |name| is not present in the pool """
        if name[0] != '!':
            try:
                obj, authobj = self.objects[name]
                if not obj or self.checkupdates:
                    self.loadmodule(name, authobj)
            except (KeyError, IndexError), err:
                authobj = None
                self.loadmodule(name, authobj)
        return self.objects[name]
    
    def loadmodule(self, name, authobj):
        """ Loads a python's module in the object path (self.path)
        The authority object is set to the |_spyro_authentication| global
        variable in the module.
        If it is not present the |authobj| argument is used instead.
        
        """
        if name.startswith('.') or os.path.sep in name:
            return
        for p in self.path:
            filename = os.path.join(p, name + '.py')
            codeloader = self.codeloader
            try:
                mtime = codeloader.getmtime(filename)
            except:
                continue
            try:
                _filename, _mtime = self.cache[name]
                if _filename == filename and _mtime == mtime: return
            except KeyError:
                pass
            lvars = ItemCodePool()
            lvars.__dict__['__builtins__'] = {}
            lvars.__dict__.update(__builtins__)
            lvars.__dict__['__builtins__']['__import__'] = self._import_
            lvars.__dict__['__name__'] = name
            # lvars.__dict__['reload'] = self._reload_
            # Double setobj operation:
            # Work around to prevent recursive import loops
            # first:
            self.setobj(name, lvars, authobj)
            exec_code(filename, self.gvars, lvars, codeloader.read(filename))
            self.cache[name] = (filename, mtime)
            try:
                auth = lvars._spyro_authentication
            except AttributeError:
                if authobj == None: # default accessor
                    auth = lvars._spyro_authentication = policies.PublicAllowCall()
                else:
                    auth = authobj
            # second and final
            self.setobj(name, lvars, auth)
            return lvars

    def delobj(self, name):
        """ Remove an object with name |name|"""
        del self.objects[name]
        try:
            del self.cache[name]
        except:
            pass
    
    def setsend(self, newid, ret, auth = None):
        """ Register an object (to send by reference)"""
        self.setobj(newid, ret, auth)
        return newid


class ModulePool(ServerObjectPool):
    """
    Register python modules as SPyRO objects. The modules must exists in a given path
    and will be appended to system path (sys.path) if the path is not in the sys.path.

    If the source code changes, the module will be reloaded.
    The packages can exists in a zip files too, or any other supported python package
    container, just giving the |statfunction| like os.stat function
    
    """
    def __init__(self, path, checkupdates=True, statfunction=os.stat):
        """ Initialize the pool """
        ServerObjectPool.__init__(self)
        self.path = []
        for x in path:
            x = os.path.abspath(x)
            self.path.append(x)
            try:
                sys.path.index(x)
            except ValueError:
                sys.path.append(x)
        self.cache = {} # objectname : timestamp
        self.statfunction = statfunction
        self.checkupdates = checkupdates
        
    def getobj(self, name):
        """ Get an object and its authorization data, it must raise a
        KeyError exception if |name| is not present in the pool """
        try:
            res = ServerObjectPool.getobj(self, name)
            if res[0] is None:
                raise KeyError("Fake")
            else:
                return res
        except KeyError, e:
            try:
                __import__(name)
            except ImportError:
                warn.warn("==================== Import error in ModulePool.getobj %s"%repr(name))
                raise e
            except SyntaxError:
                warn.warn("==================== Syntax error importing a module ModulePool.getobj %s"%repr(name))
                raise
            m = sys.modules[name]            
            abspath = os.path.abspath(m.__file__)
            if self.checkupdates:
                if abspath[-4:] in ('.pyc', '.pyo'): # .pyc -> .py
                    abspath = abspath[:-1]
                try:
                    try:
                        mtime = self.cache[name]
                        newmtime = self.statfunction(abspath).st_mtime
                        if mtime != newmtime:
                            self.cache[name] = newmtime
                            reload(m)
                    except KeyError:
                        self.cache[name] = self.statfunction(abspath).st_mtime
                except (OSError, IOError),e: # deletion of the module or package, or something with the network FS
                    raise e
            for p in self.path:
                if abspath.startswith(p):
                    try:
                        auth = m._spyro_authentication
                    except AttributeError:
                        auth = m._spyro_authentication = policies.PublicAllowCall()
                    return (m, auth)
            raise e
