#coding: utf8;
############################################################################
#                                                                          #
#    This file 'auth3k.authweb/__init__.py'                                #
#    is part of 'SPyRO: Simple Python Distributed Indexing'                #
#    Copyright (C) 2004-2005 by Eric Sadit Tellez Avila                    #
#    Copyright (C) 2006      by Eric Sadit Tellez Avila                    #
#    Copyright (C) 2007      by Eric Sadit Tellez Avila                    #
#    Copyright (C) 2008      by Eric Sadit Tellez Avila                    #
#    sadit@lsc.fie.umich.mx or donsadit@gmail.com                          #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################
"""

Authentication methods and classes for SPyRO remote objects based on remote web authenticators.
This module creates an spyro's authenticator supporting setup inline and introduces a model to
organizate access to objects and methods using users and group of users.


The authentication it's performed via user and password, or using authentication strings
(strings that are not related to user and password but grant access privilegies).
The model it's simple, and manages a lower level of access policies (see L{SPyRO.policies}). The access
permisions are granted via ACL using the property perm of a controlling Auth object.

In an Auth object, the rules of permissions are: self.perm is a dictionary (key-methodname => value-rule),
self.permobjects (Same but controls access to objects)
        
Method rules:


    - if self.perm[methodname] don't exists:
        - The method is ruled for the object rule.
        
    - if self.perm[methodname] is None:
        - The method needs a valid authentication, but not an specific user.
        
    - if self.perm[methodname] is True:
        - The method is public and everyone is allowed to execute it.

    - if self.perm[methodname] is False:
        - The method is private and cannot be accessed from SPyRO

    - if self.perm[methodname] is [ [list-of-valid-users], [list-of-valid-groups], [list-of-non-valid-users] ]
        -    Restricts the execution of the method. Note that you can use [[],[],[]] to restrict no nobody.
            The list of lists can be replaced for objects with __exists__ defined method in order to reduce the
            memory consumption in larger lists or to enable foreign dynamic access permisions.


NOTE: It's important to mention that depending of the SPyRO.codepool used, the object can be imported
like a python's module or package from any other SPyRO object, and this permisions are not checked.

"""

import os
import sys
import md5
import base64
import time
import threading
import marshal
import random
import SPyRO
from SPyRO import policies
AuthError = policies.AuthError
import warn
import types

warn = warn.Warn("auth3k.authweb")

def myxorcrypt(x,y):
    """ A simple XOR function on byte strings
    @param x: String of bytes to be XORed
    @type x: string
    @param y: String of bytes to be XORed
    @type y: string

    @return: XORed byte string
    @rtype: string
    """ 
    key = map(ord, x)
    data = map(ord, y)
    l = len(key)
    res = []
    for i,x in enumerate(data):
        data[i] = chr(x ^ key[i % l])
    return "".join(data)

try:
    from Crypto.Cipher import XOR
    def xorcrypt(x,y):
        return XOR.new(x).encrypt(y)
except ImportError:
    xorcrypt = myxorcrypt

class AuthEnviron:
    """
    Authentication Environment

    Objects of this class describes authentication expirations, seeds to generate authentication strings,
    overload of users, authentications functions, etc.
    
    Here we describe the complete behaviour of the authentication for L{Auth} objects.

    """
    def __init__(self, seedfilename=None,expirationsoft=86400,expirationhard=864000,minseeds=100,maxseeds=200):
        """

        @param expirationsoft: The maximum time than a authentication string is valid
        @param 
        expirationhard. The maximum time to create new authentication seeds
        server_auth. Array of authenticity strings
        last_auth. Last time of the server_auth
        lock. Lock for concurrent access to some methods

        If |seedfilename| is given then the random seeds are saved, and the next time 
        they will be remembered and used. If |seedfilename| is None, then volatile
        seeds will be used.
        |expirationsoft|,|expirationhard| The time (in seconds) before change auth seeds
        (soft defaults to %(expirationsoft)s and hard defaults to %(expirationhard)s)
        """%locals()
        self.minseeds = minseeds
        self.maxseeds = maxseeds
        self.expirationsoft = expirationsoft
        self.expirationhard = expirationhard
        self.server_auth = None
        self.last_auth = 0
        self.last_mtime = 0
        self.lock = threading.Lock()
        self.seedfilename = seedfilename
        self._update_global_auth()

    def _read_global_auth(self):
        if not self.seedfilename:
            return
        try:
            mtime = os.stat(self.seedfilename).st_mtime
            if self.last_mtime != mtime:
                self.last_mtime = mtime
                f = open(self.seedfilename,"r")
                self.last_auth, self.server_auth = marshal.loads(f.read())
                f.close()
        except (OSError,IOError), e:
            warn.debug("(Ignore this message if seedfilename doesn't exists yet) ", self.seedfilename, " ", e)
    
    def _random_auth(self, size):
        """ Creates a |size|-set of random strings to be used as authentication string generators """
        self.last_auth = time.time()
        self.server_auth = []
        for x in xrange(size):
            self.server_auth.append( md5.new(str((random.random(),time.time(),random.random()))).hexdigest() )
        if self.seedfilename:
            f = open(self.seedfilename,"w")
            f.write(marshal.dumps((self.last_auth, self.server_auth)))
            f.close()
            warn.debug("Writing %s seeds"%len(self.server_auth))
            try:
                os.chmod(self.seedfilename, 0600)
            except OSError, e:
                warn.debug("Can't set permissions to 0600 to the seedfile ",repr(self.seedfilename), e)
                warn.debug("Please fix this... Ignoring and running")
    
    def _update_global_auth(self):
        """ Change authorization codes """
        self._read_global_auth()
        ltime = time.time()
        if (self.last_auth == 0) or ((self.last_auth + self.expirationhard) < ltime):
            self.lock.acquire()
            try:
                self._random_auth(random.randint(self.minseeds,self.maxseeds))
            finally:
                self.lock.release()

    def getgroup(self, groupname):
        """
        Returns the list of users in |groupname|

        This method must be overloaded to overload groups. It can be overloaded
        returning an object with the method __exists__ defined (used as: user in retgroup)
        
        @param groupname: Name of the group
        @type groupname: string
        
        @return: A set representation of a group
        @rtype: An object with method __exists__ defined
        """
        try:
            import grp
            return grp.getgrnam(groupname).gr_mem
        except ImportError:
            warn.warn("Unable to load grp support, you must use another authentication method")

    def validate_user(self, user, password):
        """ Autenticate against its system's password.
        If you need to validate against any other source of passwords or validations
        this is the right method to overload.
        
        @param user: User to authenticate
        @type user: string
        @param  password: The password to use in the authentication
        @type password: string
        @return: The validated result
        @rtype: boolean
        """
        try:
            import authpam
            return authpam.authunix(user,password)
        except ImportError:
            warn.warn("Unable to load pam support, you must use other authentication method")

    def update_authentication(self, user, auth):
        """
        Updates the authentication string using an authentication string

        @param user: User associated to the authentication string
        @type user: string
        @param auth: Authentication string
        @type auth: string
        @return: The new authentication string. On failure it returns an empty string
        @rtype: string
        
        """
        if self.authenticate(user, auth, False):
            return self._create_authentication(user)
        return ""
    
    def create_authentication(self, user, password):
        """
        Creates an authentication string if the user and password are valid.
        If an error happen, it returns an empty authentication string.
        
        @param user: User associated to the authentication string
        @type user: string
        @param  password: The password to use in the authentication
        @type password: string
        @return: The authentication string. On failure it returns an empty string
        @rtype: string

        """
        if not self.validate_user(user, password):
            raise AuthError("The user %s doesn't match with the given password"%user)
        return self._create_authentication(user)
    
    def _create_authentication(self, user):
        """
        Algorithm to create the authentication string. This method don't needs
        an authentication. It's private.

        @param user: User to authenticate
        @type user: string
        @return: The authentication string
        @rtype: string
        
        """
        self._update_global_auth()
        now = str(int(time.time()))
        ran = random.randint(0, len(self.server_auth) - 1)
        dig = md5.new(now + user + self.server_auth[ran]).hexdigest()
        xor = xorcrypt(self.server_auth[0],
                       "".join(['P',dig,str(ran),',',now]))
        return base64.encodestring(xor).rstrip()

    def authenticate(self, user, auth, checksoft=True):
        """ Checks and validates the authenticity of the |auth| string for |user|
            It can update the global authentication strings
            
            @param user: User to authenticate
            @type user: string
            @param auth: Authentication string
            @type auth: string
            return: If True auth is valid for user, False otherwise.
            @rtype: boolean
        """
        if not auth: return False
        self._update_global_auth()
        try:
            authstring = base64.decodestring(auth+"\n")
            authstring = xorcrypt(self.server_auth[0], authstring)
            authtype,authdig,authtmp = authstring[0],authstring[1:33],authstring[33:]
            authran,authnow = authtmp.split(",")
            if md5.new(authnow + user + self.server_auth[int(authran)]).hexdigest() != authdig:
                return False
        except Exception, err:
            warn.warn( 'Auth.authenticate Code 1:', err)
            return False
        except:
            warn.log( 'Auth.authenticate Code 2:', err)
            return False
        if checksoft and (self.expirationsoft < (time.time() - int(authnow))):
            return False
        return True

# The default environment
DefaultAuthEnviron = None

class PublicAuthenticator:
    """ A Simple wrapper to share / export the method create_authentication """
    def __init__(self, environ):
        self.environ = environ
    
    def create_authentication(self, user, password):
        """ @see: AuthEnviron version of this method """
        return self.environ.create_authentication(user, password)

    def update_authentication(self, user, authstring):
        """ @see: AuthEnviron version of this method """
        return self.environ.update_authentication(user, authstring)

def default_public_authenticator(environ=None):
    """ Creates a public authenticator, useful to share with SPyRO """
    global DefaultAuthEnviron
    if DefaultAuthEnviron is None:
        DefaultAuthEnviron = AuthEnviron()
    if environ is None: environ = DefaultAuthEnviron
    return PublicAuthenticator(environ)

def set_default_environ(environ):
    """ Sets the default environment. The default environment is used if environ=None in the Auth constructor """
    global DefaultAuthEnviron
    DefaultAuthEnviron = environ

class AuthGood(Exception):
    """ An exception representing a successful authentication """
    pass
class AuthBad(Exception):
    """ An exception representing a failed authentication """
    pass

class Auth(policies.AuthAllowCall):
    """ This is the main class. In this object is represented all the functionality of
    access controls. It's the manager of the permisions.
    """
    def __init__(self, environ = None):
        """ Constructor an Auth object
        @param environ: Describes the environment of methods to perform authentications
        @type environ: AuthEnviron
        """
        self.perms = {  } # 'methodname' : ([list-of-valid-users],[list-of-valid-groups],[list-of-non-valid-users])
        self.permsobject = {}
        if environ is None: environ = DefaultAuthEnviron
        self.environ = environ

    def public_authenticator(self):
        """ Creates a public authentity manager """
        return PublicAuthenticator(self.environ)
        
    def _auth(self, req):
        """
        Authenticates an spyro request
        
        @param req: SPyRO request
        @type req: dict
        @return: It calls _rawauth and returns the result
        @rtype: boolean
        """
        return self._rawauth(req['auth'], req['objname'], req['attrname'])

    def _rawauth(self, auth, objname, attrname):
        """
        Authenticates a request using it's authentication tuple, object name and attribute name.
        @param auth: Authentication entity (used by _rawauth_, by default tuple (user, auth-string))
        @type auth: tuple or None
        @param objname: The name of the object that it's being accessed
        @type objname: string
        @param attrname: The name of the attribute that it's being accessed
        @type attrname: string
        @return: The interpreted result of a _rawauth_ call 
        @rtype: boolean
        """
        try:
            self._rawauth_(auth, objname, attrname)
        except AuthGood, e:
            warn.debug("Access granted as ", repr(auth), " to ", repr("%s.%s"%(objname,attrname)), ": ", e)
            return True
        except AuthBad, e:
            warn.debug("Access denied as ", repr(auth), " to ", repr("%s.%s"%(objname,attrname)), ": ", e)
            return False
        except Exception, e:
            warn.debug("Error in access denied as ", repr(auth), " to ", repr("%s.%s"%(objname,attrname)), ": ", e)
        return False

    def _rawauth_(self, auth, objname, attrname):
        """
        Authenticates a request using it's authentication tuple, object name and attribute name.
        This is the raw authentication, it raises exceptions to inform any reason of allowing or denying a
        request. It must be handled in a top call. In this method is controlled all the access: by object,
        by method, groups, etc.
        
        @param auth: Authentication entity (used by _rawauth_, by default tuple (user, auth-string))
        @type auth: tuple or None
        @param objname: The name of the object that it's being accessed
        @type objname: string
        @param attrname: The name of the attribute that it's being accessed
        @type attrname: string
        @return: Nothing, raises exceptions AuthGood and AuthBad with reasons of the decision
        @rtype: None
        """
        if auth is None:
            authuser = None
            authstring = None
        else:
            try:
                authuser, authstring = auth
            except ValueError:
                raise AuthBad, "Bad string authentication format %s"%repr(auth)
        try:
            permobj = self.permsobject[objname]
        except:
            permobj = None
        if permobj == False:
            raise AuthBad, "Object is private"
        try:
            perm = self.perms[attrname]
        except KeyError: # It is not a protected method
            perm = None
        if perm == True:
            raise AuthGood, "Method is public" # always can call
        if perm == False:
            raise AuthBad, "Method is private" # never can call
        if not self.environ.authenticate(authuser, authstring):
            raise AuthBad, "Authentication is not valid to this user"
        # 0: Allowed users
        # 1: Allowed groups
        # 2: Not allowed users
        if permobj and (authuser in permobj[2]):
            raise AuthBad, "User denied at object level"
        # Method level checking
        validuser = False
        if perm:
            if authuser in perm[2]:
                raise AuthBad, "User denied in this method"
            if authuser in perm[0]:
                raise AuthGood, "User allowed"
            for g in perm[1]:
                try:
                    if authuser in self.environ.getgroup(g):
                        raise AuthGood, "User in a valid group for this method"
                except KeyError:
                    warn.debug("_rawauth Group '", g,"' doesn't exist")
            raise AuthBad, "User not allowed by group"
        else:
            # Check if is allowed by the object
            if permobj is None:
                raise AuthGood, "Allowed at object level, general auth"
            else:
                if authuser in permobj[0]: raise AuthGood, "User allowed at object level"
                for g in permobj[1]:
                    try:
                        warn.debug(authuser, ":", self.environ.getgroup(g))
                        if authuser in self.environ.getgroup(g):
                            validuser = True
                            break
                    except KeyError:
                        warn.debug("_rawauth Group '", g ,"' doesn't exist")
            if validuser:
                raise AuthGood, "Allowed at object level"
            else:
                raise AuthBad, "User is not in a valid group for this object"
