#coding: utf8;
############################################################################
#                                                                          #
#    This file 'db3k/__init__.py'                                          #
#    is part of 'SPyRO: Simple Python Distributed Indexing'                #
#    Copyright (C) 2004-2005 by Eric Sadit Téllez Avila                    #
#    Copyright (C) 2006      by Eric Sadit Téllez Avila                    #
#    Copyright (C) 2007      by Eric Sadit Téllez Avila                    #
#    Copyright (C) 2008      by Eric Sadit Téllez Avila                    #
#    sadit@lsc.fie.umich.mx or donsadit@gmail.com                          #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################

__doc__ = """
3kong DB API 2 Interface

It saves the connection data in module space, allowing the connection to
databases using a connection name without known the data or connection details
(like database manager, users, passwords, etc).

This package simplifies the execution of sql statements, creating shortcuts to
popular processes or executions. Provides functions to create sql-injection
safe sql statements, etc.

Additionally, standardizes the way of call and behaviours of the supported managers.

db3k defines dbpools that are pools to retrieve/create connections giving the
posibility of implement persistent database connections, user or request transactions,
in a transparent way.
"""

import thread
import warn
warn = warn.Warn("db3k")
from Queue import Queue, Empty, Full

class UnknownDatabaseManager(Exception): pass

class DBData:
    """
    Storage of connection data.
    """
    def __init__(self, user, password, host, database, manager, preamble, kwargs = None, rewritesql=None):
        self.user = user
        self.password = password
        self.host = host
        self.database = database
        self.preamble = preamble
        self.rewritesql = rewritesql
        if kwargs is None: kwargs = {}
        if manager and manager.startswith('sqlite'):
            kwargs['check_same_thread'] = False
            try:
                kwargs['timeout']
            except KeyError:
                kwargs['timeout'] = 3600.0
        self.kwargs = kwargs
        if manager.startswith('sqlite'):
            try:
                from sqlite3 import dbapi2
                manager = 'sqlite3'
            except ImportError:
                try:
                    from pysqlite3 import dbapi2
                    manager = 'sqlite3'
                except ImportError:
                    try:
                        from pysqlite2 import dbapi2
                    except ImportError:
                        from sqlite2 import dbapi2
                    manager = 'sqlite2'
            self.dbm = dbapi2
        elif manager == 'mysql':
            import MySQLdb
            if self.host is None:
                self.host = '127.0.0.1'
            self.dbm = MySQLdb
        elif manager == 'postgresql':
            import pgdb
            self.dbm = pgdb
        else:
            raise UnknownDatabaseManager(manager)
        self.manager = manager
__db_infopool = { }

def set_database_data(user=None,password=None,host=None,database=None,name=None,manager='postgresql',preamble=(),kwargs=None,rewritesql=None,dbm=None):
    """
    Sets the data to connect to the database
    possible managers are:
    
    postgresql, mysql, sqlite2, sqlite3
    """
    if name is None: name = ''
    __db_infopool[name] = DBData(user,password,host,database,manager,preamble,kwargs,rewritesql)

def del_database_data(name):
    """ Removes the database connection data
    
        @param name: The name associated to the connection data
        @type name: string
        @return: Nothing, but can raise KeyError if the name was not found
    """
    del __db_infopool[name]

def get_database_data(name):
    """ Retrieves the connection data
    
        @param name: The name associated to the connection data
        @type name: string
        @return: The connection data
        @rtype: dict
    """
    return __db_infopool[name].__dict__

def create2(user=None,password=None,host=None,database=None,name=None,manager='postgresql',preamble=(),rewritesql=None,**kwargs):
    """ Creates a new connection to the database.
    
        @return: A connection and the module of the connection.
                The module is useful to manage exceptions
        @rtype: tuple (connection,module-of-connection)

        
    """
    if name is None: name = ''
    data = None
    try:
        data = __db_infopool[name]
        if database  and (data.database != database):
            raise KeyError, "KeyError Fake, database != data.database"
        if user and (data.user != user):
            raise KeyError, "KeyError Fake, user != data.user"
        if host and (data.host != host):
            raise KeyError, "KeyError Fake, host != data.host"
    except KeyError, err:
        set_database_data(user = user, password = password,
                          host = host, database = database,
                          name = name, manager  = manager,
                          preamble = preamble,
                          rewritesql = rewritesql,
                          kwargs = kwargs)
        data = __db_infopool[name]
    dbm = data.dbm
    if data.manager in ('sqlite2', 'sqlite', 'sqlite3'):
        conn = dbm.connect("%s.sqlite3"%(data.database),**data.kwargs)
    elif data.manager == 'mysql':
        conn = dbm.connect(user=data.user, passwd=data.password,
                           host=data.host, db=data.database,**data.kwargs)
    else:
        conn = dbm.connect(user=data.user, password=data.password,
                           host=data.host, database=data.database,**data.kwargs)
    cursor = conn.cursor()
    try:
        for sql in data.preamble:
            cursor.execute(sql)
    finally:
        cursor.close()
        conn.commit()
    return conn, dbm

    
def create(*args, **kwargs):
    """ The same that create2 but only returns the connection
        
        @return: A new database connection performed with associated arguments
        @rtype: database connection
    """
    return create2(*args, **kwargs)[0]

class ThreadPool:
    """ A database pool that maintains the same connection to any call in a thread.
        It mantains an implicit transaction per thread.
    """
    def __init__(self, name, manager, create_function = None):
        """
            Constructor
            
            @param name: The name associated with the connection data
            @type name: string
            @param manager: The sql manager
            @type manager: string
            @param create_function: The function to create connections (by default it uses create2)
            @type create_function: callable
            
        """
        self.name = name
        self.manager = manager
        self.connpool = {}
        self.autocommit = 0
        if create_function is None:
            create_function = create2
        self.create_function = create_function

    def delconn(self, commit=0, rollback=0):
        """
        Deletes a connection , commiting or rollbacking the thread transaction
        
        @param commit: If commit is True the transaction, default manager behaviour otherwise
        @type commit: If rollback is True the transaction is avoided. default manager behaviour otherwise.
        
        """
        conn = self.connpool.pop(thread.get_ident())
        if commit:
            conn.commit()
        else:
            if rollback:
                conn.rollback()
        conn.close()

    def getconn(self):
        """
        Retrieves a connection (the same for a whole thread, but delconn can be
        called to commit or delete and create a new connection)
        @return: A database connection
        @rtype: Database Connection 
        """
        ident = thread.get_ident()
        try:
            q = self.connpool[ident]
        except KeyError:
            q, self.dbm = self.create_function(name=self.name)
            self.connpool[ident] = q
        return q

    def putconn(self, q):
        """
        Returns to the pool a connection. By default it's ignored in this pool,
        because we use the same connection per thread.
        @param q: Database Connection
        """
        pass

    def close(self):
        """
        Closes the pool.
        """
        for q in self.connpool.values():
            q.close()
    
class QueuePool:
    """ Persistent connection pool """
    def __init__(self, name, manager):
        """
        Constructor
        
        @param name: Name of the database connection
        @type name: string
        @param manager: The name of the manager
        @type manager: string
        """
        self.name = name
        self.manager = manager
        self.queue = Queue()
        self.autocommit = 1

    def getconn(self):
        """
        Retrieves a connection, this has a queue of unused connections and get
        one of them. If there isn't connections in the queue it returns a new connection.
        
        @return: A database connection
        @rtype: Database Connection 
        """

        try:
            q = self.queue.get_nowait()
        except Empty:
            q,self.dbm = create2(name=self.name)
        return q

    def putconn(self, q):
        """
        Returns to the pool a connection. It maintains openned until the connection fails
        or the shutdown of the application.
        
        @param q: Database Connection
        @type: database connection
        
        """

        if self.manager == 'sqlite2':
            return
        try:
            self.queue.put_nowait(q)
            return
        except Full:
            pass
        q.close()

    def close(self):
        """ Closes the pool """
        try:
            while True:
                q = self.queue.get_nowait()
                q.close()
        except Empty:
            return

class Connection:
    """ Creates a new checked connection to the database manager.
    It uses the |create| function to get new connections.
    
    This class must not be used directly from a web application,
    unless you know what are you doing.
    Instead a new sub-class or managing methods must be created
    validating tables in insertions, deletions, and updates.

    """
    def __init__(self, **kwargs):
        try:
            name = kwargs['name']
        except KeyError:
            name = kwargs['name'] = ''
        try:
            pool = kwargs['pool']
            del kwargs['pool']
        except KeyError:
            pool = None
        try:
            get_database_data(name)
        except KeyError:
            set_database_data(**kwargs)
        data = get_database_data(name)
        self.dbm = data['dbm']
        self.rewritesql = data['rewritesql']
        if pool is None:
            pool = QueuePool(kwargs.get('name',''), data['manager'])
        self.pool = pool

    def getconn(self):
        """ See the specific pool getconn description """
        return self.pool.getconn()

    def putconn(self, q):
        """ See the specific pool getconn description """
        self.pool.putconn(q)
    
    def __del__(self):
        """ Deletes the object """
        self.close()
    
    def close(self):
        """
        Closes the connection
        """
        if self.pool:
            self.pool.close()
    
    def cursor(self):
        """ Returns a cursor from the RDBM backend and the connection owner of the cursor
        (cursor,connection). The cursor must be closed by the requester. The connection
        must be closed or returned to the db3k.Connection object using the self.putconn
        method
        @deprecated
        """
        try:
            conn = self.getconn()
            cursor = conn.cursor()
        except:
            warn.warn("An unexpected error was found creating the cursor. I will try to create one more time")
            try:
                conn = self.getconn()
                cursor = conn.cursor()                
            except:
                warn.warn("A second unexpected error was found creating the cursor."
                          "Please check that your RDBMs is running and working")
                raise
        return cursor, conn

    def execute(self, sqlstring, values=(), commit=1):
        """
        Execute a single cursor operation. Returns a tuple of
        resultarray, description
        
        @param sqlstring: SQL sentence
        @type sqlstring: string
        @param values: A set of values to intersect in the string
        @type values: object tuple
        @param commit: A flag to control if the connection must be commited after this sentence.
        @type commit: boolean

        @return: a tuple (resultarray, description) like (dbapi2conn.fetchall(), dbapi2conn.description)
        @rtype: tuple (list, list)

        """
        if not self.pool.autocommit:
            commit = 0
        conn = self.pool.getconn()
        cursor = conn.cursor()
        des = res = None
        dbm = self.dbm
        if self.rewritesql:
            sqlstring = self.rewritesql(sqlstring)
        try:
            # warn.debug("XXXXXXXXXXXXXXXXXXXXXXX SQL: ", sqlstring, values)
            cursor.execute(sqlstring,values)
            if cursor.description:
                res = cursor.fetchall()
                des = cursor.description
            else:
                res = ()
        except (dbm.DatabaseError, dbm.IntegrityError, dbm.NotSupportedError, dbm.ProgrammingError, dbm.InternalError, dbm.OperationalError), e:
            warn.warn("A Database error was detected." 
                      "The connection will be closed and replaced..."
                      "But the error and their consecuences will not be fixed.", e,".\n", sqlstring, values)
            cursor.close()
            if commit:
                conn.rollback()
            conn.close()
            raise
        except:
            warn.warn("An unknown error was found in db3k..."
                      "The connection will be closed and replaced..."
                      "But the error and their consecuences will not be fixed.\n", sqlstring, values)
            cursor.close()
            if commit:
                conn.rollback()
            conn.close()
            raise
        cursor.close()
        if commit:
            conn.commit()
        self.putconn(conn)
        return res, des

    def hquery(self, sqlstring, values=()):
        """
        Performs the |sqlstring| query and returns an array of dictionaries
        (fieldname => value)
        
        @param sqlstring: SQL sentence
        @type sqlstring: string
        @param values: A set of values to intersect in the string
        @type values: object tuple
        @return: a set of result records
        @rtype: list of object's dict
        
        """
        try:
            res, des = self.execute(sqlstring, values, commit=0)
        except ValueError:
            return []
        if des is None:
            return []
        names = map(lambda x: x[0], des)
        rows = []
        for r in res:
            x = {}
            rows.append(x)
            for i in range(0,len(r)):
                if isinstance(r[i], unicode):
                    x[des[i][0]] = r[i].encode('utf8')
                else:
                    x[des[i][0]] = r[i]
        return rows

    def query(self, sqlstring, values=()):
        """
        Performs the |sqlstring| query and returns an array of arrays of
        field values
        
        @param sqlstring: SQL sentence
        @type sqlstring: string
        @param values: A set of values to intersect in the string
        @type values: object tuple
        @return: a set of result records
        @rtype: list of object's list

        """
        res = self.execute(sqlstring, values)
        try:
            return self.execute(sqlstring, values, commit=0)[0]
        except IndexError:
            return []

    def ne_delete(self, tabname, condition):
        """
        Creates the SQL sentence to delete records from a table where
        that matches the given condition
        
        @param tabname: The name of the table
        @type tabname: string
        @param condition: Constraint to delete
        @type condition: dict
        @return: SQL delete sentence
        @rtype: string
        """
        if isinstance(condition, dict):
            condition = self.join_keys(condition)
        return """ DELETE FROM %s WHERE %s"""%(tabname, condition), ()

    def delete(self, tabname, condition):
        """
        Deletes an element of a table
        
        @param tabname: The name of the table
        @type tabname: string
        @param condition: Constraint to delete
        @type condition: dict
        @return: Nothing
        @rtype: None

        """
        sqlstring, values = self.ne_delete(tabname, condition)
        self.execute(sqlstring, values)

    def ne_insert(self, tabname, keyvalues):
        """ Creates the sql sentence to insert |keyvalues| (dictionary of fieldname => value) into
        the table |tabname|
        
        @param tabname: table name
        @type tabname: string
        @param keyvalues: values to insert (fieldname: value)
        @type keyvalues: dict
        @return: sql sentence
        @rtype: string

        """
        values = []
        keys = []
        if self.dbm.paramstyle == 'qmark':
            pstyle = '?'
        else:
            pstyle = '%s'

        for k,v in keyvalues.items():
            keys.append(k)
            values.append(v)
        return """ INSERT INTO %s (%s) VALUES (%s) """%(tabname, ",".join(keys), ",".join((pstyle,) * (len(values)))), values

    def insert(self, tabname, keyvalues):
        """ Insert |keyvalues| (dictionary of fieldname => value) into
        the table |tabname|
        @param tabname: table name
        @type tabname: string
        @param keyvalues: values to insert (fieldname: value)
        @type keyvalues: dict
        @return: Nothing
        @rtype: None
        
        """
        sqlstring, values = self.ne_insert(tabname, keyvalues)
        self.execute(sqlstring, values)

    def join_keys(self, keys, join_string = " AND "):
        """ Joins a condition
        @param keys: The condition expressed as pairs of key: value
        @type keys: dict
        @param join_string: string to join pairs of keys (by default AND)
        @return: SQL condition string
        @rtype: string
        """
        return join_string.join([ "%s = '%s'"%(str(k).replace("'","''"),str(v).replace("'","''")) for k,v in keys.items() ])

    def ne_update(self, tabname, newvalues, condition):
        """ Creates an sql statement to
            Update to |newvalues| (dictionary of fieldname => value)
            the table |tabname| where the |condition| is True. Condition can
            be a string or a dictionary with constraint key-pair set.
            
            @param tabname: table name
            @type tabname: string
            @param newvalues: new values to set
            @type newvalues: dict
            @param condition: a set of conjunction constraints  (keys_name = value)  pairs. 
            @type condition: dict
            @return: update sentence represented by the given arguments
            @rtype: tuple (string, values)

        """
        values = []
        string = []
        if self.dbm.paramstyle == 'qmark':
            pstyle = '?'
        else:
            pstyle = '%s'
        for k, v in newvalues.items():
            string.append("%s = %s"%(k,pstyle))
            values.append(v)
        if isinstance(condition, dict):
            condition = self.join_keys(condition)
        return """ UPDATE %s SET %s WHERE %s """%(tabname,",".join(string), condition), values

    def update(self, tabname, newvalues, condition):
        """ Update to |newvalues| (dictionary of fieldname => value)
            the table |tabname| where the |condition| is True. Condition
            can be a string or a dictionary with constraint key-pair set.
        """
        sqlstring, values = self.ne_update(tabname, newvalues, condition)
        self.execute(sqlstring, values)

def index_result(qres, key):
    """ Returns a dictionary indexed by |key| of the hquery results (an integer
        if the results are from Connection.query or a string if they are results of
        Connection.hquery). Duplicated keys are not supported.
        
        @param qres: query result
        @type qres: list's list or dict's list
        @param key: Index or Key to index 
        @type key: int or string
        @return: dictionary indexed by key
        @rtype: dict of records
        
        """
    res = {}
    for q in qres:
        res[q[key]] = q
    return res

def index_result_dup(qres, key):
    """ Returns a dictionary indexed by |key| of thehquery results (an integer
        if the results are from Connection.query or an string if they
        are results of Connection.hquery).
        Duplicated keys are supported using a list as container.

        @param qres: query result
        @type qres: list's list or dict's list
        @param key: Index or Key to index 
        @type key: int or string
        @return: dictionary indexed by key
        @rtype: dict of records

        """
    res = {}
    for q in qres:
        try:
            res[q[key]].append(q)
        except KeyError:
            res[q[key]] = [q]
    return res

if __name__ == '__main__':
    import getpass
    set_database_data(user='sadit',password=getpass.getpass(),database='catalogos')
    conn = Connection()

    print conn.query("""
    SELECT conkey, confkey,
          (SELECT relname FROM pg_class WHERE pg_class.oid = conrelid) as "InitialTable",
          (SELECT relname FROM pg_class WHERE pg_class.oid = confrelid) as "FinalTable"
    FROM pg_constraint
    WHERE contype = 'f' AND conrelid = (SELECT oid FROM pg_class WHERE relname = 'comment')""");
    conn.close()
