#coding: utf8;
"""
An asynchat based http client, similar in api structure to httplib.

Working with asynchat means working with callbacks, then some callbacks must
be given to manage the http responses.

"""
import warn
import asyncore
import asynchat
import socket
import re
_head_sep = re.compile(":\s*")

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

class HTTPResponse:
    """
    An http response
    """
    def __init__(self, headers, respfile):
        """
        Constructor
        
        @param headers: Headers of the response
        @param respfile: A file like data object
        """
        self.headers = headers
        self.respfile = respfile
        respfile.seek(0)

    def read(self, size=-1):
        """
        Reads size bytes of the respfile
        """
        return self.respfile.read(size)

    def getheaders(self):
        """
        Return all the headers
        """
        return self.headers
    
    def getheader(self, name):
        """
        Return one header value
        @param name: The name of the header to be retrieved
        """
        return _getheader(name, self.headers)

    def close(self):
        """
        Close the response
        """
        pass

def _getheader(name, headers, default=""):
    """
    Internal function to retrieve a header
    """
    name = name.lower()
    for n,v in headers:
        if name == n:
            return v
    return default

class HTTPConnection(asynchat.async_chat):
    """
    HTTP client
    """
    def __init__(self, host, port=80, outobj=None, begincallback=None, endcallback=None):
        """
        Constructor
        
        @param host: The host or ip of the remote server
        @param port: Port of the HTTP server
        @param outobj: An object with the write method (L{file} like write capabilities)
        @param begincallback: Callback to handle body connections
        @param endcallback: Calback to handle a closing connection
        
        """
        asynchat.async_chat.__init__(self)
        self.address = (host, port)
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.headerline = None
        self.set_terminator("\n")
        self.responseline = None
        self.isbody = False
        self.headers = []
        self.connect(self.address)
        self.buffer = ""
        self.chunkstat = 0
        if outobj is None:
            outobj = StringIO()
        self.outobj = outobj
        self.begincallback = begincallback
        self.endcallback = endcallback
        
    def _dowrite(self, string):
        """
        Perform a write. Internal handler
        
        @param string: string to write
        """
        self.outobj.write(string)

    def _doendcallback(self):
        """
        Default close handler
        """
        if self.endcallback:
            self.endcallback(self)
            self.endcallback = None 

    def _dobegincallback(self):
        """
        Default begin body handler
        """
        if self.begincallback:
            self.begincallback(self)
            self.begincallback = None

    def _getbuffer(self):
        """
        Returns the buffer
        """
        d = self.buffer
        self.buffer = ""
        return d
    
    def handle_connect(self):
        """
        Handle a connection
        """
        pass

    def handle_close(self):
        """
        Handle close connection
        """
        self._doendcallback()
        self.close()
    
    def collect_incoming_data(self, data):
        """
        Save a remote connection
        """
        if self.isbody:
            if self.chunkstat:
                if self.chunkstat == 1:
                    self.buffer = self.buffer + data
                elif self.chunkstat == 2:
                    self._dowrite(data)
                elif self.chunkstat == 3:
                    pass
            else:
                self._dowrite(data)
        else:
            self.headerline = data.rstrip()

    def getresponse(self):
        """
        Return an http response
        """
        return HTTPResponse(self.headers, self.outobj)
    
    def found_terminator(self):
        """
        Found a terminator handler
        """
        if self.isbody:
            if not self.chunkstat:
                self._doendcallback()
                return
            if self.chunkstat == 1: # read chunk size
                self.chunkstat += 1
                size = int(self._getbuffer().rstrip(),16)
                self.chunksize = size
                self.set_terminator(size)
                return
            if self.chunkstat == 2: # read chunk
                self.chunkstat += 1
                self.set_terminator("\n") # to read the last chunk
                return
            if self.chunkstat == 3: # read last line and return the stat
                self.chunkstat = 1
                self.set_terminator('\n')
                if self.chunksize == 0:
                    self._doendcallback()
                return
        if self.headerline == '':
            if self.chunkstat:
                self.set_terminator("\n")
            else:
                l = _getheader("content-length", self.headers)
                try:
                    l = int(l)
                except (TypeError, ValueError):
                    self.set_terminator(None)
                else:
                    self.set_terminator(l)
            self.headerline = None
            self.isbody = True
            self._dobegincallback()
        else:
            if not self.responseline:
                self.responseline = self.headerline
                self.command, self.path, self.version = self.responseline.split(None,2)
            else:
                try:
                    name, value = _head_sep.split(self.headerline,1)
                except ValueError, err:
                    warn.warn("HEADER ERROR:",repr(self.headerline),err)
                    raise
                name = name.lower()
                if name == 'transfer-encoding':
                    if value == 'chunked':
                        self.chunkstat = 1
                self.headers.append((name, value))
    
    def putheader(self, key, value):
        """
        Put an http header (key: value)
        @param key: header's name
        @param value: header's value
        """
        self.push("%s: %s\r\n"%(key.title(), str(value)))
    
    def putrequest(self, method, url, version="HTTP/1.0"):
        """
        Put the request
        @param method: Method (e.g. GET, POST, etc)
        @param url: Location of the request (e.g. /index.html)
        @param version: HTTP version (e.g. HTTP/1.0)
        """
        self.push("%s %s %s\r\n"%(method, url, version))

    def endheaders(self):
        """
        Finalizes headers sending and empty new line
        """
        self.push("\r\n")
    
if __name__ == '__main__':
    # xhost = 'barrapunto.com'
    # xhost = 'osnews.com'
    xhost = 'localhost'
    # "www.httpwatch.com/httpgallery/chunked/"
    xhost = "www.httpwatch.com"
    xlocation="/"
    xlocation="/httpgallery/chunked/"
    xport = 8000
    xport = 80
    import sys
    xconn = HTTPConnection(xhost,xport, outobj=sys.stderr, endcallback=lambda conn: warn.debug("callback:",conn))
    xconn.putrequest("GET",xlocation)
    xconn.putheader("Host",xhost)
    xconn.endheaders()
    asyncore.loop()
    if False:
        resp = xconn.getresponse()
        print "=============================="
        print xconn.command, xconn.path, xconn.version
        print resp.getheaders()
        print resp.read()
    
