#coding: utf8;
############################################################################
#                                                                          #
#    This file 'shttp/http.py'                                             #
#    is part of 'SPyRO: Simple Python Distributed Indexing'                #
#    Copyright (C) 2004-2005 by Eric Sadit Tellez Avila                    #
#    Copyright (C) 2006      by Eric Sadit Tellez Avila                    #
#    Copyright (C) 2007      by Eric Sadit Tellez Avila                    #
#    Copyright (C) 2008      by Eric Sadit Tellez Avila                    #
#    sadit@lsc.fie.umich.mx or donsadit@gmail.com                          #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################
"""
This module controls the HTTP backend for SPyRO and the embedded http server
(currently medusa) of SPyRO/Dweba.

We can find in this module:
    - Handler to call spyro using http web services
    - Handler to call spyro using http cgi
    - Handler to call any kind of python functions
    - Identify where we must compress the output of functions
    - Custom responses from spyro
    
"""
try:
    from cStringIO  import StringIO
except ImportError:
    from StringIO import StringIO

import sys
import md5
import string
import re
import cgi
import gzip
import types
from medusa import http_server
from medusa import producers
from medusa import default_handler
from medusa import filesys
import warn
import stat
import mimetypes
mimetypes.init()

import thread
import threading
__thread_data__ = {}

def getrequest():
    """ Returns the spyro's request dictionary that
    rules the current thread or shttp.http request. If this thread is not
    ruled by an http request, it will raise a KeyError exception"""
    return __thread_data__[thread.get_ident()][0]

def getthreaddata():
    """
        Returns all the data related to the thread
    """
    return __thread_data__[thread.get_ident()]

def setheader(header):
    """
        Set an http header to the response of this thread
    """
    __thread_data__[thread.get_ident()][1].append(header)

def _setrequest(data):
    """ Internal function to set the data to be stored by calling thread
    """
    headers = []
    __thread_data__[thread.get_ident()] = [data, headers]

def _delrequest():
    """ Internal function to removes the data stored by calling thread
    """
    del __thread_data__[thread.get_ident()]


class HttpCgiError(Exception):
    """
        Represents an exception from an Http CGI
    """
    def __init__(self, *args):
        self.args = args
        
class HttpMethodError(HttpCgiError):
    """
        Exception in an Http called method
    """
    pass

class CustomResponse:
    def __init__(self, data, headers={}, content_type='text/plain', status=200, reason="Response Follows", options={}, content_type_from_path=None):
        """
        This class is used as an exception to send Custom Responses from a
        SPyRO method to the HTTP web server. The web server must known how 
        handle this exception.
        
        @param data: The data to be send in the response (an str object)
        @param headers: A dictionary with pair valued headers
        @param content_type: The content type of the data
        @param status: The number of status to present in the response (if it's posible).
                        The default value is 200 (OK).
        @param reason: Reason response line
        @param options: Options to response
        @param content_type_from_path: If this argument is given, it must be a
                        filename and the content type will be guessed from this filename.
        """
        self.reason = reason
        self.options = options
        self.status = status
        self.data = data
        self.headers = headers
        self.content_type_from_path = content_type_from_path
        self.content_type = content_type

def getenviron(request):
    """
    Prepares a dictionary environment with CGI/1.1 environment variables
    
    @param request: The http request
    @type request: http handler object
    @return: The environment
    @rtype: dict
    """
    env = {}
    try:
        length = int(request.get_header('content-length'))
    except TypeError:
        length = 0
    # print "================", "url:", request.uri
    for h in request.header:
        # print "DDDDDDDDD", h
        k,v = h.split(": ",1)
        env["HTTP_%s"%(k.replace("-","_").upper())] = v
    if length:
        env['CONTENT_LENGTH'] = length
    path, params, query, fragment = request.split_uri()
    if query is None:
        query = ''
    else:
        query = query[1:]
    env['QUERY_STRING'] = query
    # print dir(request)
    env['REQUEST_METHOD'] = request.command
    # useful for spyro and other simple handlers
    # if self.simple_environ: return env
    env['SERVER_SOFTWARE'] = http_server.VERSION_STRING
    env['SERVER_PROTOCOL'] = request.version
    env['SERVER_NAME'] = request.channel.server.addr[0] #self.server.server_name
    env['SERVER_PORT'] = str(request.channel.server.addr[1])
    env['GATEWAY_INTERFACE'] = 'CGI/1.1'
    # uqrest = urllib.unquote(rest)
    env['PATH_INFO'] = path[len(request.uri_handler):]
    env['SCRIPT_NAME'] = path.rsplit('/',1)[-1]
    # env['PATH_TRANSLATED'] = 'No-Available' # self.translate_path(uqrest)
    #host = self.address_string()
    #if host != self.client_address[0]:
    env['REMOTE_HOST'] = env['REMOTE_ADDR'] = request.channel.addr[0]
    authorization = request.get_header("authorization")
    if authorization:
        authorization = authorization.split()
        if len(authorization) == 2:
            import base64, binascii
            env['AUTH_TYPE'] = authorization[0]
            if authorization[0].lower() == "basic":
                try:
                    authorization = base64.decodestring(authorization[1])
                except binascii.Error:
                    pass
                else:
                    authorization = authorization.split(':')
                    if len(authorization) == 2:
                        env['REMOTE_USER'] = authorization[0]
    # ===================================================================0
    # Headers multiples
    for c in ('cookie', 'accept', 'content-type'):
        s = request.get_header(c)
        if s is None:
            s = ''
        env[c.upper().replace('-','_')] = s
    #
    # ===================================================================0
    
    # XXX Other HTTP_* headers
    # Since we're setting the env in the parent, provide empty
    # values to override previously set values
    for k in ('QUERY_STRING', 'REMOTE_HOST', 'CONTENT_LENGTH',
              'HTTP_USER_AGENT', 'HTTP_COOKIE'):
        env.setdefault(k, "")
    # print "=================================== ENV: ", env
    return env

import Cookie, urllib
def getcookiespyroauth(authcookiename,environment):
    """
    Retrieves the cookies that authenticates SPyRO calls
    
    @param authcookiename: The name of the cookie
    @type authcookiename: string
    @param environment: CGI/1.1 Environment
    @type environment: dict
    @return: tuple representing the auth entity to call spyro methods
    @rtype: list or tuple
    """
    try:
        c = environment['HTTP_COOKIE']
        cookies = Cookie.SimpleCookie()
        cookies.load(c)
        values = urllib.unquote(cookies[authcookiename].value)
        values = values.split(',')
        x = [ urllib.unquote(cookies[v].value) for v in values ]
        return x
    except KeyError:
        return None

def webspyro(spyroserver, httpreq, environment, returntype="text/plain", authcookiename='SPyRO_AuthMethod', mode="simple"):
    """
    CGI interface to call spyro methods
    
    @param spyroserver: The spyro server
    @param httpreq: The http request object
    @param environment: The CGI environment
    @param returntype: The return type of the response
    @param authcookiename: The cookie name to retrive the authentication entity
    @param mode: The url parsing-mode, in other words, how the url must be interpreted.
                The available modes:
                    - 'simple' (default). The url is a object-name/method-name.
                        Arguments can be sent using key-value pairs using GET or POST.
                    - 'arg'. It uses the last part of the location as first argument of
                        the method, as string. The method is specified in 'simple' format with
                        the remaining left-location. 
                        Arguments can be sent using key-value pairs using GET or POST.
    """
    try:
        path = environment['PATH_INFO'].lstrip("/")
        if not path: raise KeyError
    except KeyError:
        path = 'index'
    # parts = [ environment['SCRIPT_NAME'].split('/')[-1] ]
    parts = path.split('/')
    if mode == 'arg':
        # parts.extend( path.split("/") )
        args = (parts.pop(),)
        attrname = parts.pop()
        objname = ".".join(parts)
    else:
        args = ()
        attrname = parts.pop()
        objname = ".".join(parts)
    kwargs = getkwargs(httpreq.input_data, environment)
    auth = getcookiespyroauth(authcookiename,environment)
    # print (objname, attrname, (), kwargs, auth)
    res = spyroserver.handle_request(objname, attrname, args, kwargs, auth, reqtype='CALL')
    if res['errcode'] != 0:
        raise HttpMethodError(res)
    webmethodwrite(httpreq, res['retvalue'], returntype)

def httpcallspyro(spyroserver, returntype='text/plain', authcookiename='SPyRO_AuthMethod',  mode="simple"):
    """
    Creates a lexical closure to execute a function with
    query_string and post requests as function arguments
    
    @param spyroserver: The SPyRO server to the call

    If the proper authorization policies aren't disabled (policy 'can_follow_objname'):
    C{
    /SPyRO/Chart/Bar/Draw
    /SPyRO -> Registered URL base to spyroserver
    /Chart -> Object
    /Bar -> A property of /Chart
    /Draw -> A method of /Bar
    }    
    @param returntype: The return type of the function to be placed as
    Content-Type header

    Note: If httpcall is used as decorator the declared function
    will be replaced, use regcall of http object as decorator.
    
    """
    def directcall(httpreq, environment):
        webspyro(spyroserver, httpreq, environment, returntype, authcookiename, mode)
    return directcall

def handlespyro(spyroserver, content_type="application/xml", overture=None, rewrite=None):
    """
    Handle spyro request. Direct call using spyro's formats
    
    @param spyroserver: The spyro server
    @param content_type: The result content type
    @param overture: A function to protect the request against indeseable client.
            This function will be executed before than intepretation of the spyro's
            message and execution of the instructions.
    """
    def directcall(httpreq, environment):
        res = spyroserver.handle_http(httpreq.input_data, environment["QUERY_STRING"], overture, rewrite)
        webmethodwrite(httpreq, res, content_type)
    return directcall
    
def _getkwargs_qs(querystring):
    """
    Retrieves keywords from query string.
    
    @param querystring: The QUERY_STRING environment or an string with the same meaning
    @return: dictionary of keywords
    """
    res = {}
    for k,v in cgi.parse_qsl(querystring):
        try:
            r = res[k]
            try:
                r.append(v)
            except AttributeError:
                r = [r, v]
            res[k] = r
        except KeyError:
            res[k] = v
    return res

    
def getkwargs(data,environment):
    """
    Retrieves the keyword arguments encoded in the HTTP request (POST and GET)
    
    @param data: The POST data (If POST)
    @param environment: CGI environment
    @return: Dictionary of keywords in the request
    """
    if environment['REQUEST_METHOD'].lower() == 'post':
        kwargs = _getkwargs_qs(environment['QUERY_STRING'])
    else:
        return _getkwargs_qs(environment['QUERY_STRING'])    
    fields = cgi.FieldStorage(fp=StringIO(data),environ=environment)
    for x in fields.keys():
        isfile=None
        try:
            isfile = fields[x].filename
        except AttributeError:
            isfile = None
        except:
            warn.debug("=========== http keywords:", repr(x))
            warn.warn("XXXXXXXXXXXXX KEYWORD ERRORS:", fields)
            raise
        if isfile is None:
            kwargs[x] = fields.getvalue(x)
        else:
            kwargs[x] = fields[x]
    return kwargs

def webmethodwrite(httpreq, res, returntype, status=200, headers=(), options={}):
    """
    This is the main writter of functions and methods called from the web (like spyro's calls)
    It's a very flexible tool. If it's possible it tries  to compress the output to enhance the
    network transmision performance.
    
    @param httpreq: The http request
    @param res: Response of the method
    @param returntype: The return type
    @param status: The status code of the response
    @param headers: can be one of the following:
        - A dictionary
        - A tuple/list of tuple/list of two elements
        - A tuple/list of a string per entry
        - A mixin of the previous two
    @param options: A dictionary to simplify the configuration and extensibility of the function.
    """
    httpreq.reply_code = status
    httpreq["Content-Type"] = returntype
    env,_headers = getthreaddata() # getrequest()
    # print env
    try:
        res.read
        isfile = True
    except AttributeError,e:
        isfile = False
        if not isinstance(res, types.StringTypes):
            res = str(res)
    try:
        if isfile:
            raise KeyError("Refuse to compress a direct file")
        try:
            encodings = env['HTTP_ACCEPT_ENCODING'].split(",")
        except KeyError:
            pass
        else:
            for enc in encodings:
                enc = enc.strip()
                if not enc.endswith("gzip"):
                    continue
                string = StringIO()
                gzfile = gzip.GzipFile(mode="wb", fileobj = string)
                gzfile.write(res)
                # gzfile.flush()
                gzfile.close()
                res = string.getvalue()
                httpreq["Content-Encoding"] = enc
                break
        httpreq["Content-Length"] = "%d"%len(res)
    except KeyError:
        pass
    except (ValueError,TypeError), err:
        warn.warn(err)
        pass
    try:
        items = headers.items()
    except AttributeError:
        items = headers
    _headers.extend(items)
    _headers.append(("Connection","close"))
    for item, value in _headers:
        httpreq[item] = str(value)
    if isfile:
        httpreq.push(producers.file_producer(res))
    else:
        httpreq.push(producers.scanning_producer(res))
    # not used, we prefeer to call after _delrequest() to close anything attached
    # to this request before to try to send data
    # TODO: Real solution, make the work outside of this thread (maybe a listening socket to create select events)
    httpreq.done()
    
def webmethod(httpreq, environment, method, returntype="text/plain"):
    """
    Calls a registered function using an urihandler
    
    @param httpreq: The http request
    @param environment: The CGI environment
    @param method: The method to be called, a callable object
    @param returntype: The return type
    @return: Nothing
    """
    kwargs = getkwargs(httpreq.input_data, environment)
    res = method(**kwargs)
    webmethodwrite(httpreq, res, returntype)

class uri_handler:
    """
    A class to represent uri handlers. This class manages handlers associated with an uri
    
    """
    def __init__(self, location, handler, rawcall, virtual_host, content_type, rewrite, spawnthread):
        """
        Constructor
        
        @param location: Location (uri) associated to the handler
        @param handler: The callable handler
        @param rawcall: If True then the call will be in raw format (see below)
        @param virtual_host: The virtual host name (must match with the host
                    header, regexp in string format)
        @param content_type: The default content type of the response
        @param rewrite: The rewrite function to be applied to the URI
        @param spawnthread: If True then it will spawn a new thread to solve the request
        """
        self.location = location
        self.handler = handler
        self.virtual_host = re.compile(virtual_host)
        self.content_type = content_type
        self.rawcall = rawcall
        self.rewrite = rewrite
        self.spawnthread = spawnthread
        
    def match(self, request):
        """
        Checks if the request must be handled for this uri handler
        
        @param request: http request
        @return: True if this request must responsed by this handler, False otherwise
        """
        if self.virtual_host.search(request.get_header("host")):
            try:
                request._rewrite_done
            except AttributeError:
                if self.rewrite:
                    self.rewrite(request)
                request._rewrite_done = 1
            if request.uri.startswith(self.location):
                request.uri_handler = self.location
                return 1
        return 0
    
    def handle_request(self, request):
        """
        Handle one request
        @param request: The http request
        @returns: Nothing
        """
        (path, params, query, fragment) = request.split_uri()
        if request.command == 'POST':
            request.collector = collector (self, request)
        else:
            self.continue_request('', request)

    def continue_request(self, data, request):
        """
        This method solves the request after collecting the POST input. It spawns
        a new thread if self.spawnthread is True.
        
        @param data: Data of the POST request (if any)
        @param request: The http request
        @returns: Nothing
        """
        if self.spawnthread:
            thread.start_new_thread(self.solve, (data, request))
        else:
            self.solve(data, request)

    def solve(self, data, request):
        """
        This method solves the request called from continue_request (in the same thread
        or in a new thread). It handles many exceptions
        
        @param data: Data of the POST request (if any)
        @param request: The http request
        @returns: Nothing

        """
        try:
            try:
                self._solve(data, request)
            except:
                type, value, traceback = sys.exc_info()
                if isinstance(value, HttpMethodError):
                    error = repr(value.args)
                elif isinstance(value, Exception):
                    try:
                        error = repr(value.message)
                    except AttributeError:
                        error = repr(value)
                else:
                    error = repr(value)
                if True:
                    weberror = error
                else:
                    weberror = "Private error, please contact with your system administrator"
                warn.warn("Internal Error %s,  Exception: "%repr(getrequest()), error)
                webmethodwrite(request, 'Internal Server Error: %s'%error, 'text/plain',status=500)
        finally:
            _delrequest()

    
    def _solve(self, data, request):
        """
        This method solves the request called from self.solve. It doesn't handles any exception
        (except for CustomResponses that are customized outputs, not errors).
        
        @param data: Data of the POST request (if any)
        @param request: The http request
        @returns: Nothing
        """
        environ = getenviron(request)
        _setrequest(environ)
        try:
            # generate response
            if self.rawcall:
                request.input_data = data
                self.handler(request, environ)
            else:
                kwargs = getkwargs(data, environ)
                data = self.handler(**kwargs)
                webmethodwrite(request, data, self.content_type)
        except CustomResponse, resp:
            if resp.content_type_from_path:
                resp.content_type = mimetypes.guess_type(resp.content_type_from_path)[0]
            webmethodwrite(request, resp.data, resp.content_type, resp.status, resp.headers, options=resp.options)

class collector:
    "gathers input for POST and PUT requests"
    # TODO: Prevent buffer out of memory in collector
    def __init__ (self, handler, request):

        self.handler = handler
        self.request = request
        self.data = ''

        # make sure there's a content-length header
        cl = request.get_header ('content-length')

        if not cl:
            request.error (411)
        else:
            cl = string.atoi (cl)
            # using a 'numeric' terminator
            self.request.channel.set_terminator (cl)

    def collect_incoming_data (self, data):
        """
        Collects data from the sockets.
        """
        self.data = self.data + data

    def found_terminator (self):
        """
        Handler of string terminator
        """
        # set the terminator back to the default
        self.request.channel.set_terminator ('\r\n\r\n')
        self.handler.continue_request (self.data, self.request)

class filesystem(filesys.os_filesystem):
    """
    Listing of directories
    """
    def longify(self, (path, stat_info)):
        """
        List directories
        """
        # print dir(stat_info)
        if stat.S_ISDIR(stat_info.st_mode):
            ftype = "DIR"
            suffix = '/'
        else:
            ftype = "FILE"
            suffix = ''
        return """<tr><td>[%s]</td><td>%0.1f</td><td>KiB</td><td><a href="%s">%s%s</a></td></tr>"""%(ftype,stat_info.st_size/1024.0,path,path,suffix)

class HTTPChannel(http_server.http_channel):
    """
    An HTTP channel  overwritting the medusa one to support multithreading.
    """
    def __init__(self, channel, conn, addr):
        """
        Constructor
        
        @param channel: Original channel
        @param conn: Connection
        @param addr: Remote address (client address)
        @type addr: tuple (ip, port)
        """
        self.sendlock = threading.Lock()
        self.must_close = 0
        http_server.http_channel.__init__(self, channel, conn, addr)
           
    def close (self):
        """
        Closes the channel
        """
        if self.sendlock.locked():
            self.must_close = 1
            # we can close it right now (there's a bug in python's multithreading I/O)
            # but we can remove it from the list of the select or poll function, after
            # this we can remove using the must_close flag
            http_server.http_channel.del_channel(self)
            return
        self.sendlock.acquire()
        try:
            http_server.http_channel.close(self)
        finally:
            self.sendlock.release()
    
    def handle_error(self):
        self.close()
    
    def initiate_send (self):
        """
        Internal method to initialize data sending and inform to the main thread
        that data is present.
        """
        if self.sendlock.locked():
            return
        self.sendlock.acquire()
        try:
            http_server.http_channel.initiate_send(self)
        finally:
            self.sendlock.release()
            if self.must_close:
                # http_server.http_channel.del_channel(self)
                self.close() # removing the connection, lattely
            else:
                NotifySentinel(9998)

class HTTPServer(http_server.http_server):
    """
    The http server specialization using medusa http_server as base http server.
    """
    def __init__(self,
                 server_address, # A tuple with the server address
                 document_root=".", # The document root
                 ):
        """
        Constructor
        @param server_address: tuple (ip, port) to attach the server
        @param document_root: Path to the document root
        """
        # ThreadingTCPServer.__init__(self, server_address, HandlerClass)
        self.document_root = document_root
        http_server.http_server.channel_class = HTTPChannel        
        dh = default_handler.default_handler(filesystem(document_root))
        http_server.http_server.__init__(self, server_address[0], server_address[1])
        self.install_handler(dh)
    
    def register_urihandler(self, uri, handler, virtual_host='', rawcall=1, content_type="text/html", rewrite=None, spawnthread=1):
        """ Set an urihandler
        @param uri: The URI to be handled
        @param handler: a function that handles the event
        @param virtual_host: The name of the virtual host, default to '' (default virtual host)
            - Prototype:
                - function(httpreq, environment)
                - httpreq is an HTTPRequestHandler Object
        @param rawcall: The handler is raw if rawcall is True
        @param content_type: Default content type for the handler
        @param rewrite: A rewrite function in order to match
        @param spawnthread: If It's True a new thread is created to call the handler. By default it's true.
        """
        self.install_handler(uri_handler(uri, handler, rawcall, virtual_host, content_type, rewrite, spawnthread))

    def unregister_urihandler(self, uri, virtual_host=''):
        """
        Deletes the uri-handler associated to |uri|
        @param virtual_host: The name of the virtual host, default to '' (default virtual host)
        """
        rem = []
        for h in self.handlers:
            if h.match(uri):
                rem.append(h)
        for h in rem:
            self.remove_handler(h)

    def regobj(self, uri, object, returntype="text/plain", authority=None):
        """
        Register an object to handle requests
        
        @param uri: The associated uri
        @param object: The object handler
        @param returntype: The default return type of results
        @param authority: Not implemented
        """
        def handler(httpreq, environment):
            res = httpreq.cgi_info[1]
            method = object
            for x in res.split('?')[0].split('/'):
                if x == '': continue
                method = getattr(method,x)
            webmethod(httpreq, environment, method, returntype)
        self.register_urihandler(uri, handler)
        return object

    def regcall(self, uri, returntype='text/plain'):
        """ Register uri |uri| to be handled by a function
        This method can be used safelly as decorator
        
        Example
        @httpobj.regcall('/selected-uri','text/html')
        def square(x):
            int(x)
            return x*x
        """
        def decorator(function):
            self.register_urihandler(uri, httpcall(function, returntype))
            return function
        return decorator

import asyncore
import socket

class Sentinel(asyncore.dispatcher_with_send):
    """
    Internal Sentinal system to inform asyncore changes from threads. Client version.
    """
    lock=threading.Lock()
    def __init__(self, ip, port, conn=None):
        self.out_buffer=''
        if conn is None:
            # self.name = "DESDE CLIENTE"
            if Sentinel.lock.locked():
                return
            Sentinel.lock.acquire()
            try:
                asyncore.dispatcher.__init__ (self)
                self.create_socket (socket.AF_INET, socket.SOCK_STREAM)
                self.connect((ip, port))
                # self.close()
            finally:
                Sentinel.lock.release()
        else:
            # self.name = "DESDE SERVER"
            asyncore.dispatcher.__init__(self, conn)
        self.closed = 0
    
    def handle_connect(self):
        # print "S handle_connect", self.name
        pass
    
    def handle_write(self):
        # print "S handle_write", self.name
        pass

    def handle_read(self):
        # print "S handle_read: ", self.name
        self.recv(10)
        pass
    
    def handle_error(self):
        # print "S handle_close", self.name
        self.close()
        self.closed = 1

class AsyncSentinel(asyncore.dispatcher):
    """
    Internal Sentinal system to inform asyncore changes from threads. Server version.
    """
    def __init__ (self, ip, port):
        asyncore.dispatcher.__init__ (self)
        self.create_socket (socket.AF_INET, socket.SOCK_STREAM)
        self.set_reuse_addr()
        self.bind ((ip, port))
        # lower this to 5 if your OS complains
        self.listen (1024)
        self.closed = 0

    def writable (self):
        return 1

    def handle_read (self):
        # print "AS handle_read"
        self.recv(10)
    
    def handle_write (self):
        # print "AS handle_write"
        pass
    
    def readable (self):
        return 1

    def handle_connect (self):
        # print "AS handle_connect"
        pass

    def handle_accept (self):
        # print "AS handle_accept"
        conn, addr = self.accept()
        Sentinel(None, None, conn)
        pass

    def handle_error(self):
        # print "AS handle_error"
        self.close()
        self.closed = 1
    
_ASent = None
_CSent = None

def NotifySentinel(SentinelPort=10101, SentinelHost='127.0.0.1'):
    """ Function to notify server sentinel of changes using the a client """
    global _ASent, _CSent
    while 1:
        if _ASent is None:
            _ASent = AsyncSentinel(SentinelHost, SentinelPort)
            _CSent = Sentinel(SentinelHost, SentinelPort)
        if _CSent.closed or _ASent.closed:
            _ASent = _CSent = None
        else:
            _CSent.send("A") # wake up!
            break

if __name__ == '__main__':
    import asyncore
    import asynchat
    s = HTTPServer(('localhost', 8000))
    def showrequest(req, env):
        return repr(( req, env)).replace("<","&lt;").replace(">","&gt;")

    def square(x):
        x = float(x)
        return x * x
    
    s.register_urihandler("/call/show", showrequest)
    s.register_urihandler("/call/square", square, rawcall=0)
    asyncore.loop()
    
