#coding: utf8;
############################################################################
#                                                                          #
#    This file 'template3k/__init__.py'                                    #
#    is part of 'dom3k'                                                    #
#    Copyright (C) 2004-2005 by Eric Sadit Téllez Avila                    #
#    Copyright (C) 2006      by Eric Sadit Téllez Avila                    #
#    Copyright (C) 2007      by Eric Sadit Téllez Avila                    #
#    Copyright (C) 2008      by Eric Sadit Téllez Avila                    #
#    sadit@lsc.fie.umich.mx or donsadit@gmail.com                          #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################

__doc__ = """
This module provides support for simple but powerful templates. 
template3k is a template regular language based on substitution of
commands for their evaluation.

There exists two levels of execution:

Static level (substitution on template compilation time) or dynamic substitution
(executed on running time).

The entire template a python living code, so you can define python classes or
functions inside and use them at any time in the template.

It's based on commands that looks like simple html tags (to be easily used with
graphical html design applications).  The templates could not be an xml document,
but the commands will be named in the same way.

A template is executed with the method 'merge(values)' where values is a dictionary
with values to be used by the commands. Functions and lambdas in dynamic substitutions
receives as argument the merge's dictvalues as a dictionary

Commands:

python:keyword
   <python:keyword id="Name" />
       Replaces the command with values["Name"]
   <python:keyword id="Name">Default text</python:keyword>
       Replaces the command with values["Name"], with optionally default text

python:exec
   <python:exec id="Name">
   ...Python Code...
   </python:exec>
       Replaces the command with the result of compute the python code inside the command
       and whose varname is Name.
       If Name variable is a function or lambda it will be a dynamic substitution

python:code
   <python:code id="Name">
   ...Python Code...
   </python:code>
       Replaces the command with an empty string but the code is saved in the environment of
       the template, so its definitions could be used later

python:include
   <python:include id="Name">TemplatePath</python:include>
       Replaces the command with the template substituion of TemplatePath template. If must exists an special
       keyword |template_pool| in the merge dictionary to be used to get the template. If this key is not
       given in the merge, the default module get_template will be used. The arguments in the merge will be
       passed to this new template merge.
       
python:eval
   <python:eval id="Name">...Python Expression...</python:eval>
       Replaces the command with the result of the evaluation of the Python Expression inside the command tags.
       If Name is send to merge it will be replaced for the merge argument.
       If the result of the expression is a function or a lambda, it will be a dynamic substitution, a string
       otherwise. If Name is send to merge, the expression will not be executed. There is a implicit argument
       in the environment named dictvalues, is not needed in the Python Expression, but can be used in the
       Python Expression as well.

python:call
   <python:call id="Name">...Python Expression...</python:call>
       Replaces the command with the result of the evaluation of the Python Expression inside the command tags.
       If Name is send to merge it will be replaced for the merge argument.
       This expression will be always executed. If Name is send to merge, the expression will not be executed.

python:locale
   <python:locale id="Objname.Name" />
   <python:locale id="Objname.Name">Default Text</python:locale>
       Replaces the tag with the locale translation of defined as "Objname.Name". If there is not locale for
       Objname.Name the text inside the tags are used, or it must be mandatory if there is a self closed tag.
       There is a closed relation with the spyro-object locale, which provides a graphical locale editor, and
       an export and import locale tool.

       If there is not defitions of Objname.Name in the locale database, it will be inserted automatically
       in the moment of template compilation.
       The locale is taken from HTTP_ACCEPT_LANGUAGE environment variable.
"""

import re
import warn
import os
import types

_pool = None
_template_pattern_full=re.compile(r"""
     <(?P<startend> /)?\s*python:(?P<command>\w+) 
     (\s* id\s*=\s*["'](?P<id>[^"']+)["'])?
     \s* (?P<endself> /)? \>
     """, re.IGNORECASE | re.DOTALL | re.MULTILINE | re.VERBOSE )

class UnknownTemplateCommand(Exception): pass
class SystaxTemplateError(Exception): pass

class Template:
    def __init__(self, name, data):
        self.name = name
        self.vars = {}
        self.template = self._compiletemplate(data)

    def _compilecommand(self, comm, xid, args, xprog):
        if comm == 'keyword':
            if len(args) == 0:
                xprog.append(lambda dv: dv[xid])
            elif len(args) == 1:
                xprog.append(lambda dv: dv.get(xid, args[0]))
            else:
                xprog.append(lambda dv: dv.get(xid, lambda xdv: self.evaltemplate(dv, args)))
        elif comm in ('dynamic-locale', 'locale'):
            import locale3k
            objname, name = xid.rsplit(".", 1)
            lenargs = len(args)
            if lenargs == 0:
                xdef = None
            else:
                assert lenargs == 1, "locale commands only receives strings as arguments %s"%args
                xdef = eval(args[0])                    
            def getlocale(dd={}):
                return dd.get(xid, locale3k.getname(objname, name, locale3k.getweblocale(), xdef))
            if comm == 'locale':
                xprog.append(getlocale())
            else:
                xprog.append(getlocale)
        elif comm in ('exec','code'):
            assert len(args) == 1, "%s command only receives strings as arguments %s"%(comm,args)
            c = compile(args[0]+"\n", self.name + ':id:' + xid,'exec')
            exec c in self.vars, self.vars
            if comm == 'code':
                xprog.append('')
            else:
                xprog.append(lambda dv: dv.get(xid, self.vars[xid]))
        elif comm == 'eval':
            assert len(args) == 1, "%s command only receives strings as arguments %s"%(comm,args)
            p = eval(args[0], self.vars, self.vars)
            xprog.append(lambda dv: dv.get(xid, p))
        elif comm == 'call':
            p = eval("lambda dictvalues: (%s)"%args[0], self.vars, self.vars)
            xprog.append(lambda dv: dv.get(xid, p))
        elif comm == 'include':
            assert len(args) == 1, "%s command only receives strings as arguments %s"%(comm,args)
            def include_template(dictvalues):
                try:
                    g = dictvalues['template_pool'].get_template
                except KeyError:
                    g = get_template
                return g(args[0]).merge(dictvalues)
            xprog.append( include_template )
        else:
            raise UnknownTemplateCommand('Unknown Template Command %s'%repr(comm))

    def _compiletemplate(self, data):
        xprog=[]
        while 1:
            d = _template_pattern_full.search(data)
            if d is None:
                xprog.append(data)
                break
            xstart, xend = d.span()
            xprog.append(data[0:xstart])
            x = d.groupdict()
            comm = x['command']
            args = []
            xid = x['id']
            if x['startend']:
                index = -1
                while 1:
                    try:
                        p = xprog[index]
                    except IndexError, e:
                        warn.debug("*** Error in template: dumping template text:\n", ",\n".join(map(repr,xprog)),"\n")
                        raise Exception((repr(("Closing tag don't match with any openned tag", comm, xid, self.name))))
                    if (type(p) is types.TupleType) and (p[0] == comm):
                        xid = p[1]
                        args = xprog[index+1:]
                        xprog[index:] = ()
                        break
                    index -= 1
                # warn.debug("PRECL: comm: %s, xid: %s>"%(comm,xid), xprog)
                # warn.debug( "args: ", (len(args), args) )
                self._compilecommand(comm, xid, args, xprog)
                # warn.debug("FINAL: comm: %s, xid: %s>"%(comm,xid), xprog)
            elif x['endself']:
                self._compilecommand(comm, xid, args, xprog)
            else:
                assert xid, "id attribute must be set in %s:%s, or your are closing with an openning tag"%(comm,xid)
                xprog.append((comm, xid))
            data = data[xend:]
        self.xprog = xprog

    def evaltemplate(self, dictvalues, prog):
        """ Merge the template with a set of values |dictvalues| """
        res = []
        for p in prog:
            while callable(p):
                p = p(dictvalues)
            if type(p) in types.StringTypes:
                res.append(p)
            else:
                res.append(str(p))
        return "".join(res)
    
    def merge(self, dictvalues=None, **kwargs):
        if dictvalues is None:
            dictvalues = kwargs
        return self.evaltemplate(dictvalues, self.xprog)
        
class Pool:
    def __init__(self, path="."):
        self.path = path
        self._CACHE = {} # name => mtime, template

    def get_template(self, template, data=None):
        _CACHE = self._CACHE
        filename = os.path.join(self.path,template)
        if data is None:
            try:
                s = os.stat(filename)
            except OSError:
                warn.debug("Template %s don't exists"%repr(filename))
                raise
            try:
                _mtime, _temp = _CACHE[template]
                if _mtime == s.st_mtime:
                    return _temp
            except KeyError:
                pass
            f = file(filename)
            data = f.read()
            f.close()
            mtime = s.st_mtime
        else:
            mtime = 0
        c = Template(filename, data)
        _CACHE[template] = (mtime, c)
        return c

def get_template(temname, data=None):
    global _pool
    if _pool is None:
        _pool = Pool()
    return _pool.get_template(temname, data)
