#coding: iso-8859-1;
############################################################################
#                                                                          #
#    This file 'warn/__init__.py'                                          #
#    is part of 'SPyRO: Simple Python Remote Objects'                      #
#    Copyright (C) 2004,2005 by Eric Sadit Téllez Avila                    #
#    Copyright (C) 2006,2007 by Eric Sadit Téllez Avila                    #
#    Copyright (C) 2008      by Eric Sadit Téllez Avila                    #
#    sadit@lsc.fie.umich.mx or donsadit@gmail.com                          #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #

#    sadit@lsc.fie.umich.mx or donsadit@gmail.com                          #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #

#    sadit@lsc.fie.umich.mx or donsadit@gmail.com                          #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################


__all__ = ['debug','log','warn','stack','halt']
""" Debug and warning system (just error display). """

# from SPyRO import SPyRO
import traceback
import sys
import time
import os

class FileWriter:
    def __init__(self, name, outfile=None):
        self.name = name
        if outfile:
            self.file = outfile
        else:
            self.file = file(name,'a+b',0)

    def write(self, string, msgtype=None):
        self.file.write(string)
    
class SyslogWriter:
    def __init__(self, name):
        self.name = name

    def write(self, string, msgtype=None):
        syslog.openlog(self.name)
        syslog.syslog(syslog.LOG_ERR, string)
        syslog.closelog()


# In output is stored every special handler to manage the output.
# The handlers must support the write method, and must
# receive a string as argument.

output = {}
def logger_simple(logname, string, msgtype):
    """ A simple output to the logname log """
    try:
        Warn.output[logname].write(string)
    except KeyError:
        sys.stderr.write(string)
    
def logger_create_by_function(logid, logsplitter, openfun=None, hookfun=None):
    """ Creates a logger function (a closure)
    With |logid| name (to the set the output filename) and a function that choose the
    a tuple of log names to be used by call (|logsplitter|). The output files will be opened
    using the function |openfun|(logid, logname) and must return a file like object
    (only write method is needed)
    if |openfun| is None every file is opened in append and binary mode and
        output filenames are ${main-log-directory}/log.${logid}.${logname}
        Where ${main-log-directory} can be set with the function set_main_log_directory
           ${logid} It the logid gived to this function
           ${logname} It's the returned value of the logsplitter function
    if |hookfun| is given, then this function is called after every warn call
        and it can be used to any external treatment of the warn, using the following prototype
        def hookfun(string, msgtype, logslist)

        Where msgtype can be given in every warn call as a keyword (msgtype=TYPE) or the type of warn if is not
        given (i.e. a strng between HALT, DEBUG, LOG, WARN, and STACK). This can be used to email certain types
        of messages or any other treatment of some warning messages.
        logslist is the list of logs that will be used as output, this can be used to decide behaviours too.
    ------------------------
    For example:
        logger = logger_create_by_function("by-date", lambda x,y: (time.strftime("%y-%m"),))
        logger = logger_create_by_function("spyro-user", lambda x,y: (spyro.getrequest()['auth'][0],)) # Using the SPyRO's Module
        logger = logger_create_by_function("spyro-obj", lambda x,y: (spyro.getrequest()['objname'],))# Using the SPyRO's Module
        logger = logger_create_by_function("log-ip", lambda x,y: (shttp.http.getrequest()['HTTP_REMOTE_ADDR'],)) # Using the SPyRO's shttp Module
        set_main_log_function(logger)

    The exact prototype of |logsplitter|
    iterableobject logsplitter(str logid, str logname)

    Note: The opened files are never closed, if you need to close them, use the
    dictionary warn.Warn.outputs

    key: logname (not logid), value: output file

    ------------------------------
    If  logsplitter raises an exception (like KeyError for unknown logs or a failed log name resolution)
    the default log is used (logger_simple)
    """
    if openfun is None:
        def xopenfun(_logid, _logname):
            xfilename = os.path.join(_MAIN_LOG_DIRECTORY,"log.%s.%s"%(_logid,_logname))
            return file(xfilename, "ab", 0)
        openfun = xopenfun
    def _thelogger(_logname, string, msgtype):
        try:
            lognamelist = logsplitter(logid, _logname)
            for logname in lognamelist:
                try:
                    outfile = Warn.output[logname]
                except KeyError:
                    outfile = Warn.output[logname] = openfun(logid, logname)
                outfile.write(string)
            if hookfun:
                hookfun(string, msgtype, lognamelist)
        except Exception:
            logger_simple(_logname, string, msgtype)
            if hookfun:
                hookfun(string, msgtype, [_logname])
    return _thelogger

_MAIN_LOG_FUNCTION = logger_simple
_MAIN_LOG_DIRECTORY = "." # In the same directory of the application

def set_main_log_directory(dirname):
    """ The main log directory. For default loggers """
    global _MAIN_LOG_DIRECTORY
    _MAIN_LOG_DIRECTORY = dirname

def set_main_log_function(logger):
    """ Initialices the global logger, this is the core of the warn module
    And it can be used to perform fancy splitted logs, like log by session,
    user, module, spyro objects, by network address, by time, by day, etc.
    
    |logger| The function logger. It can be any object that can be called.
    The prototype must be:
    
    void logger(logname, logdata, msgtype)

        Where logname is the name of the log
              logdata is the data to be logged
              msgtype is a string in ("LOG","HALT","WARN","DEBUG", "STACK")
                      meaning the type of message or can be given at any
                      call to warn methods and can be any other identifier
    """
    global _MAIN_LOG_FUNCTION
    _MAIN_LOG_FUNCTION = logger

class Warn:
    LOG_SYSLOG = 1
    LOG_FILE = 2
    LOG_STDERR = 3
    output = output
    def __init__(self, name, outfile=None, filename=None, logtype=None):
        """ Creates a Warn object. Warn objects provide useful
        tools to show errors and warnings.
        |name| The name of the warn object
        |outfile| An output object in case that you have a default output.
        If None, a file will be created
        |logtype| If it must use syslog. If none it uses the default logger
        Possible values are: Warn.LOG_FILE, Warn.LOG_SYSLOG, Warn.LOG_STDERR, or None
        Default: None
        If LOG_FILE is specified the constructor needs any of the
        following keyword arguments:
        filename='filename-to-output' or outfile=fileobject

        NOTE: If output already has a handler, this handler is used overriding
        any of the opening instructions. In order to override, you need
        to modify the handler (using warn.output) directly
        """
        self.name = name
        #### To avoid an overload of a previously defined handler
        if Warn.output.has_key(name):
            return
        if logtype == Warn.LOG_STDERR:
            Warn.output[name] = sys.stderr
        elif logtype == Warn.LOG_SYSLOG:
            Warn.output[name] = SyslogWriter(name)
        elif logtype == Warn.LOG_FILE:
            if not outfile:
                if not filename:
                    raise ValueError("You must specify a filename if you want a LOG_FILE Warn object")
                outfile = open(filename,'a+b',0)
            Warn.output[name] = outfile
        elif logtype is None:
            pass
        else:
            raise ValueError("Invalid type of Warn")
    
    def write(self, string, msgtype=None, useformat=True):
        """ Output for loggin, appends the current time, the name of the log, and the pid """
        if msgtype is None:
            msgtype = "LOG"
        if useformat:
            self._write("%s %s: %s\n"%(time.ctime(time.time()), self.name, string), msgtype)
        else:
            self._write(string, msgtype)

    def _write(self, string, msgtype):
        """ Direct write to the output """
        _MAIN_LOG_FUNCTION(self.name, string, msgtype)
    
    def debug(self, *args, **kwargs):
        """ Shows the args on the stderr """
        self.write("".join(map(str,args)) + "\n", kwargs.get("msgtype","DEBUG"))
        
    def log(self, *args, **kwargs):
        """ Shows the args on a log """
        self.write("".join(map(str,args)) + "\n", kwargs.get("msgtype","LOG"))
    
    def warn(self,*args, **kwargs):
        """ Shows |args| as warning, and print the last exception catched """
        exc=sys.exc_info()
        self.write("".join(("".join(map(str,args)), "\n%s: %s"%(exc[0],exc[1]))), kwargs.get("msgtype","WARN"))
        traceback.print_tb(exc[2],file=self)
 
    def stack(self,*args, **kwargs):
        """ Shows |args| as warning, and print the stack of current caller """
        self.write("".join(map(str,args)) + "\n", kwargs.get("msgtype","STACK"))
        traceback.print_stack(file=self)

    def halt(self,*args, **kwargs):
        """ Displays |args|, print the last_exception and exits with code 127"""
        exc=sys.exc_info()
        self.write("".join(("".join(map(str,args)), "\n", str(exc[1]))), kwargs.get("msgtype","HALT"))
        traceback.print_tb(exc[2],file=self)
        sys.exit(1)

_name = "app-%s"%sys.argv[0]

_w = Warn(_name)
debug = _w.debug
log = _w.log
warn = _w.warn
stack = _w.stack
halt = _w.halt
write = _w.write
