#coding: utf8;

__doc__ = 'DB Easyform creator'
import auth3k
import auth3k.authweb as authweb
_spyro_authentication = authweb.Auth()
_spyro_authentication.permsobject['Guigo.easyform'] = ((),(),())

import os
import re
import time
import db3k
import easyform
from easyform import Entity, Group
import warn

class BaseUI:
    def __init__(self, objname):
        self.objname = objname
    
    def populate(self, destdb, internal_fields=()):
        for tab in self.show_tables():
            tab = tab[0]
            foreigns = self.foreigns_fields(tab)
            for desc in self.describe_table(tab):
                d = {'objname': self.objname,
                     'form': tab,
                     'name': desc[0],
                     'type_code': desc[1],
                     'play_size': desc[2],
                     'internal_size': desc[3],
                     'precision': desc[4],
                     'scale': desc[5],
                     'null_ok': desc[6],
                     'ui_ok': not ( (desc[0] in internal_fields) or ("%s.%s"%(tab,desc[0]) in internal_fields) )
                     }
                try:
                    foreigntable, foreignfield = foreigns[desc[0]]
                    d['foreigntable'] = foreigntable
                    d['foreignfield'] = foreignfield
                except KeyError:
                    pass
                # warn.debug("DDDDDDDDDDDD: ", d)
                destdb.insert('admin_guigo',d)

class Api2UI(BaseUI):
    def __init__(self, objname, conn, dbname, schema, structure = {}):
        BaseUI.__init__(self, objname)
        self.conn = conn
        self.dbname = dbname
        self.schema = schema
        self.structure = structure # table => foreigns
        # self.conn.execute("connect information_schema")

    def show_tables(self):
        # return self.conn.execute("select table_name,table_catalog,table_rows,table_schema from tables")
        return self.structure.values()

    def describe_table(self,tabname):
        return self.conn.execute("select * from %s"%tabname)[1]

    def foreigns_fields(self,tabname):
        return self.structure[tabname]

class MysqlUI(BaseUI):
    def __init__(self, objname, conn, dbname, schema):
        BaseUI.__init__(self, objname)
        self.conn = conn
        self.dbname = dbname
        self.schema = schema
        # self.conn.execute("connect information_schema")

    def show_tables(self):
        # return self.conn.execute("select table_name,table_catalog,table_rows,table_schema from tables")
        return self.conn.query("select table_name from tables where table_schema = '%s'"%self.dbname)

    def describe_table(self,tabname):
        x = self.conn.query("""
        select column_name, data_type, NULL, NULL, NULL, NULL, is_nullable = 'YES'
        FROM columns where  table_schema = %s and table_name = %s""", (self.dbname, tabname))
        return x

    def foreigns_fields(self,tabname):
        # return []
        tabs = self.describe_table(tabname)
        res = {}
        for d in tabs:
            field = d[0]
            if '_' in field:
                foreign = field.split("_",1)
                res['field'] = (foreign[0],foreign[1])
        return res
    
class PgUI(BaseUI):
    def __init__(self, objname, conn, database, schema):
        BaseUI.__init__(self, objname)
        self.conn = conn
        self.database = database
        self.schema = schema

    def show_tables(self):
        return self.conn.query("SELECT tablename FROM pg_catalog.pg_tables WHERE schemaname = '%s'"%self.schema)
    
    def describe_table(self, tabname):
        return self.conn.execute("SELECT * FROM %s.%s LIMIT 1"%(self.schema, tabname))[1]
    
    def foreigns_fields(self,tabname):
        foreigns = self.conn.query("""
    SELECT
        (SELECT relname FROM pg_class WHERE pg_class.oid = conrelid) as "localtable",
        (SELECT attname FROM pg_attribute WHERE attrelid = conrelid AND attnum = conkey[1]) as "localfield",
        (SELECT relname FROM pg_class WHERE pg_class.oid = confrelid) as "foreigntable",
        (SELECT attname FROM pg_attribute WHERE attrelid = confrelid AND attnum = confkey[1]) as "foreignfield"
    FROM pg_constraint
    WHERE contype = 'f' AND conrelid = (SELECT oid FROM pg_class WHERE relname = '%s' and relnamespace = (SELECT oid FROM pg_namespace WHERE nspname = '%s'))"""%(tabname,self.schema))
        res = {}
        for localtable, localfield, foreigntable, foreignfield in foreigns:
            res[localfield] = (foreigntable, foreignfield)
        return res


class SqliteUI(BaseUI):
    def __init__(self, objname, conn, database, schema):
        BaseUI.__init__(self, objname)
        self.conn = conn
        self.database = database
        self.schema = schema
        self.foreigns = None
    
    def show_tables(self):
        x = self.conn.query("SELECT DISTINCT tbl_name FROM sqlite_master WHERE type='table'")
        return x
    
    def describe_table(self, tabname):
        if self.foreigns is None:
            self.foreigns = {}
        self.foreigns[tabname] = {}
        qres = self.conn.execute("SELECT * FROM %s LIMIT 1"%(tabname))[1]
        qres = [ list(x) for x in qres ]
        master = self.conn.hquery("SELECT * FROM sqlite_master WHERE type='table' AND tbl_name = ?",(tabname,))[0]
        tablecomp = re.compile(r"create \s+ table \s+ \w+ \s* \( ( .+ ) \)",re.I|re.X|re.S)
        # fieldcomp = re.compile(r"([\w]+)?\s*([\(,;])?\s*",re.I|re.X|re.S)
        exprcomp = re.compile(r"""(?P<name> \w+) \s+ """
                              r"""(?P<type> \w+ (?P<size> \s* \(\d+\) )? ) \s*"""
                              r"""(?P<refereces> references \s+ (?P<tabname> \w+) \s* \( (?P<fields> [^\)]+) \) )?"""
                              r"""(?P<null_ok>[^,]+)?,?""",re.I|re.X|re.S)
        m = tablecomp.search(re.sub(r"--.*?[\n]","",master['sql']))
        xtab = {}
        for item in exprcomp.findall(m.group(1)):
            xname, xtype, xsize, xref, xreftab, xreffields, xnullok = map(str.lower, item)
            xsize = xsize.replace("(","").replace(" ","").replace(")","")
            xtab[xname] = {'xname': xname,
                           'xtype': xtype.split('(')[0],
                           'xsize': xsize,
                           'xref': xref,
                           'xreftab': xreftab,
                           'xreffields': xreffields,
                           'xnullok': xnullok,
                           }
        
        for q in qres:
            name = q[0].lower()
            print "==================>",xtab.keys()
            info = xtab[name]
            if q[1] is None:
                q[1] = info['xtype']
            if q[2] is None:
                q[2] = info['xsize']
            if q[3] is None:
                q[3] = info['xsize']
            if q[-1] is None:
                q[-1] = not("not" in info['xnullok'])
            if info['xref'] != '':
                for field in info['xreffields'].split(","):
                    field = field.strip()
                    self.foreigns[tabname][name] = (info['xreftab'], field)
        return qres
    
    def foreigns_fields(self,tabname):
        if self.foreigns is None:
            self.describe_table(tabname)
        try:
            self.foreigns[tabname]
        except KeyError:
            self.describe_table(tabname)
        return self.foreigns[tabname]

class ProjectUI:
    def __init__(self, objname, guigodb):
        self.objname = objname
        self.guigodb = guigodb

    def get_config(self):
        return self.guigodb.hquery("SELECT * FROM admin_guigo_project WHERE objname = ? ORDER BY id",(self.objname,))[0]

    def get_internal_fields(self, tabname):
        qres = self.guigodb.query("SELECT name FROM admin_guigo WHERE objname = ? AND form = ? AND ui_ok = 0 ORDER BY id",(self.objname,tabname))
        return [ q[0] for q in qres ]

    def show_tables(self):
        # return self.conn.execute("select table_name,table_catalog,table_rows,table_schema from tables")
        qres = self.guigodb.query("SELECT DISTINCT form FROM admin_guigo WHERE objname = ? ORDER BY id",(self.objname,))
        return [ str(q[0]) for q in qres ]
    
    def describe_table(self,tabname):
        qres = self.guigodb.query("SELECT name, type_code, play_size, internal_size, precision, scale, null_ok "
                                  "FROM admin_guigo WHERE objname = ? and form = ? ORDER BY id",(self.objname,tabname))
        res = []
        for q in qres:
            xq = list(q)
            xq[0] = str(q[0])
            xq[1] = str(q[1])
            res.append(xq)
        return res
    
    def foreigns_fields(self,tabname):
        return self.guigodb.query("""SELECT form, name, foreigntable, foreignfield """
                                  """FROM admin_guigo WHERE objname = ? AND form = ? AND """
                                  """foreigntable IS NOT NULL""",(self.objname,tabname))

    def foreigns_fields_dictionary(self, tabname):
        res = {}
        for form, name, fortab, forfield in self.foreigns_fields(tabname):
            res[name] = (fortab, forfield)
        return res
            
def import_user_interface(guigodb,
                          database=None,
                          objname=None,
                          user=None,
                          password=None,
                          host=None,
                          manager='postgresql',
                          language='en',
                          schema=None,
                          internal_fields=''):
    dbname = "%s%s"%(database, time.time())
    if manager == 'postgresql':
        conn = db3k.Connection(name=dbname,database=database,user=user,password=password,host=host,manager=manager)
        # warn.debug(locals())
        dbm = PgUI(objname, conn, database, schema)
    elif manager == 'mysql':
        conn = db3k.Connection(name=dbname,database='information_schema',user=user,password=password,host=host,manager=manager)
        dbm = MysqlUI(objname, conn, database, schema)
    elif manager.startswith('sqlite'):
        conn = db3k.Connection(name=dbname,database=database,user=user,password=password,host=host,manager=manager)
        dbm = SqliteUI(objname, conn, database, schema)
    else:
        raise Exception("This manager (%s) is not supported for UI generation"%manager)
    dbm.populate(guigodb,map(str.strip,internal_fields.split(",")))

class UserInterfaceGenerator:
    def __init__(self, guigodb, objname):
        self.guigodb = guigodb
        self.objname = objname
        self.connui = ProjectUI(objname, guigodb)
        
    def normalizetype(self, type, length=0):
        """Returns a tuple easyform-type"""
        if   type.find('char') >= 0:
            return 'string'
        elif type.find('int')  >= 0:
            return 'int'
        elif type.find('text') >= 0:
            return 'text'
        elif (type.find('float') >= 0) or (type == 'numeric'):
            return 'float'
        elif type.find('list') >= 0:
            return 'list'
        elif type.find('datetime') >= 0 or type.find('timestamp') >= 0:
            return 'datetime'
        elif type.find('time') >= 0:
            return 'time'
        elif type.find('date') >= 0:
            return 'date'
        elif type.find('bool') >= 0:
            return 'boolean'
        else:
            raise TypeError, "Unknown database type %s"%type

    def groupfromtable(self, tabname):
        """ Creates a group from a table """
        des = self.connui.describe_table(tabname)
        foreigns = self.connui.foreigns_fields(tabname)
        config = self.connui.get_config()
        dfor = {}
        for f in foreigns:
            dfor[str(f[1]).lower()] = f
        onsubmit=r"""<python:keyword id="__onsubmit__" />"""
        g = Group(tabname,onsubmit=onsubmit,language=config['language'],internal_fields=self.connui.get_internal_fields(tabname))
        for d in des:
            name, type_code, play_size, internal_size, precision, scale, null_ok = d
            try:
                name = str(name).lower()
                # warn.debug("OOOO %s>"%tabname, repr(name), dfor)
                localtable, localfield, foreigntable, foreignfield = dfor[name]
                # warn.debug("================================> ",name, dfor[name])
                markuptemplate = (
                                  """<select onfocus="callto({'rawhandler': createhandlerselect('%%(value)s'), 'handlerargs': this},"""
                                  """'<python:keyword id="__objname__" />','preview_%(foreigntable)s',['%(foreignfield)s'])" name="%%(name)s:%%(type)s" """
                                  """ id="select-%(foreigntable)s_%(foreignfield)s_<python:keyword id='__id__' />" """
                                  """>"""
                                  """</select>\n"""
                                  """<a style="text-decoration: none;" """
                                  """href="javascript: DOM3K_replaceListOptions([['','']],'select-%(foreigntable)s_%(foreignfield)s_<python:keyword id="__id__" />'); void(0);">"""
                                  """<python:locale id="actions.clear" /></a>"""
                                  """<python:keyword id="__editmode__">\n"""
                                  """\n/\n"""
                                  """<a style="text-decoration: none;" """
                                  """href="javascript: showformmodify('<python:keyword id="__objname__" />', '%(foreigntable)s'); void(0);">"""
                                  """<python:locale id="actions.new" /></a>"""
                                  """</python:keyword>\n"""
                                  """\n/\n"""
                                  """<a style="text-decoration: none;" href="javascript: callto({'showresults': """
                                  """function(d) {DOM3K_createFrame(d,'title=\\\\'<python:locale id="actions.search_form" />\\\\' ');}}, """
                                  """ '<python:keyword id="__objname__" />', 'form_search_%(foreigntable)s', """
                                  """['choose_search','select-%(foreigntable)s_%(foreignfield)s_<python:keyword id="__id__" />']); void(0);">"""
                                  """<python:locale id="actions.search" /></a>"""
                                  """<python:keyword id="__fulltextsearch__">\n"""
                                  """\n/\n"""
                                  """<a style="text-decoration: none;" href="javascript: callto({'showresults': """
                                  """function(d) {DOM3K_createFrame(d,'title=\\\\'<python:locale id="actions.search_form" />\\\\' ');}}, """
                                  """ '<python:keyword id="__objname__" />', 'form_search_fulltext_%(foreigntable)s', """
                                  """['choose_search','select-%(foreigntable)s_%(foreignfield)s_<python:keyword id="__id__" />']); void(0);">"""
                                  """<python:locale id="actions.search_fulltext" /></a>"""
                                  """</python:keyword>\n"""
                                  """\n"""%locals())
            except KeyError:
                markuptemplate = None
            constraints = []
            if not null_ok:
                constraints.append("mandatory")
            entity = Entity(name=name, type=self.normalizetype(type_code), constraints=constraints, markuptemplate=markuptemplate)
            g.append( entity )
            # print name, type_code, play_size, internal_size, precision, scale, null_ok
        e = Entity("",type="submit",language=config['language'])
        e.value = """<python:locale id="general.submit" />"""
        g.append(e)
        return g
    
    def createsite(self, dbconn, objpath):
        """ Creates the complete site
        |name| The name of the site.
        |outdir| The output directory.
        """
        tabres = self.connui.show_tables()
        config = self.connui.get_config()
        pycont = easyform.PythonContainer()
        pycont.begin()
        for tabname in tabres:
            g = self.groupfromtable(tabname)
            pycont.append(g.name, g)
        import packadmin
        try:
            localname = os.path.join(objpath,self.objname)
            config['fullobjpath'] = "/".join((objpath,self.objname))
            if os.path.isfile(localname + ".py"):
                raise Exception("The full path to this object %(fullobjpath)s already exists, please remove it first to create"%config)
            config['fullobjname'] = packadmin.get_objname_from_path(config['fullobjpath'])
        except packadmin.PathNotAllowed, e:
            warn.warn("create site: ",e)
            raise
        pycont.end(config,self.connui)
