#coding: utf8;
import re
import appconfig
import db3k
import math
import warn
import spyropack

SQLDATA = """
BEGIN;

CREATE TABLE ft_doc_%(name)s
(      	
	id integer primary key,
	cond varchar(32),
	maxfreq integer
);

CREATE TABLE ft_voc_%(name)s
(
	word varchar(16),
	-- must we use domain right here to get fast searches?
	docid integer references ft_doc(id),
	freq integer
);

CREATE TABLE ft_params_%(name)s
(
       name varchar(16) UNIQUE,
	-- must we use domain right here to get fast searches?
       value integer
);

CREATE INDEX ftvwrd_%(name)s ON ft_voc_%(name)s(word);
CREATE INDEX ftvdoc_%(name)s ON ft_voc_%(name)s(docid);
CREATE INDEX ftvcnd_%(name)s ON ft_doc_%(name)s(cond);

COMMIT;
"""

class sqlindex:
    def __init__(self, name, conn, wordreg=r"[a-zA-Z0-9]{2,16}"):
        self.wordreg = re.compile(wordreg)
        self.name = name
        self.conn = conn
        self.create_index()
        try:
            self.vectorize = appconfig.vectorize
        except AttributeError:
            self.vectorize = self._vectorize
    
    def create_index(self):
        name = self.name
        exists = len(self.conn.query("SELECT 1 FROM sqlite_master WHERE tbl_name = ? AND type='table'",("ft_doc_%s"%name,)))
        if not exists:
            spyropack.sql_script_data(self.conn, SQLDATA%locals())
        
    def drop_index(self):
        self.conn.execute("DROP TABLE ft_voc_%s"%self.name)
        self.conn.execute("DROP TABLE ft_doc_%s"%self.name)
        self.conn.execute("DROP TABLE ft_params_%s"%self.name)
    
    def handle_update(self, data, dataid):
        docid = self.get_docid(dataid)
        self.conn.delete("ft_voc_" + self.name, {'docid': docid})
        self.insert_document(data, docid)

    
    def handle_insert(self, data, dataid):
        docid = self.get_docid(dataid)
        self.insert_document(data, docid)
        
    def handle_delete(self, data, dataid):
        docid = self.get_docid(dataid)
        name = self.name
        self.conn.delete("ft_voc_" + self.name, {'docid': docid})
        self.conn.delete("ft_doc_" + self.name, {'id': docid})
        vec, maxfreq = self.vectorize(data)
        for word,freq in vec.items():
            self.conn.execute("""UPDATE ft_params_%s SET value = value - ? WHERE name = ?"""%name,(freq,word))

    def get_sqlinfo(self, cond):
        """ returns the original dataid """
        # return docid.split('/',1)
        return cond
    
    def get_docid(self, dataid):
        """ returns an integer identifing the document """
        # cond = "%s/%s"%(domain, dataid)
        cond = dataid
        for x in xrange(2):
            res = self.conn.query("SELECT id FROM ft_doc_%s WHERE cond = ?"%self.name, (cond,))
            try:
                return res[0][0]
            except IndexError:
                pass
            self.conn.execute("INSERT INTO ft_doc_%s (cond, maxfreq) VALUES(?,?)"%self.name,(cond,0))
        raise Exception("Database error, get_docid ", locals())

    def _vectorize(self, doc):
        res = {}
        maxfreq = 0
        for w in self.wordreg.findall(doc):
            w = w.lower()
            try:
                res[w] = freq = res[w] + 1
            except KeyError:
                freq = res[w] = 1
            maxfreq = max(maxfreq, freq)
        return res, maxfreq
    
    def insert_document(self, doc, docid):
        conn = self.conn
        vec, maxfreq = self.vectorize(doc)
        name = self.name
        for w,f in vec.items():
            self.conn.insert("ft_voc_" + name,{'word': w, 'docid': docid, 'freq': f})
            if 0 < len(self.conn.query("""SELECT 1 FROM ft_params_%s WHERE name = ? LIMIT 1"""%name, (w,))):
                self.conn.execute("""UPDATE ft_params_%s SET value = value + ? WHERE name = ?"""%name,(1,w))
            else:
                self.conn.insert("ft_params_" + name, {'value': 1, 'name': w})
        self.conn.update('ft_doc_' + name, {'maxfreq': maxfreq}, {'id': docid})
        return maxfreq
    
    def search(self, query, size=10, cmpop="LIKE", qtype="vector"):
        assert cmpop in ('LIKE','=',">","<","<>","<=",">="), "search comparator is not supported %s"%repr(like)
        name = self.name
        conn = self.conn
        try:
            ndocs = conn.query("""SELECT id FROM ft_doc_%s ORDER BY id DESC LIMIT 1"""%name)[0][0]
        except IndexError:
            return []
        qvec, qmaxfreq = self.vectorize(query)
        qmaxfreq = float(qmaxfreq)
        # words = ",".join(map(repr, qvec.keys()))
        qres = []
        maxsize = size * 3
        docres = {}
        for w,qfreq in qvec.items():
            if cmpop == 'LIKE': # due to sqlite3 limitations we need to avoid like operator and use a work around
                wb = w[:-1] + chr(ord(w[-1]) + 1)
                wordcmp = "(word >= '%s' AND word < '%s')"%(w, wb)
            else:
                wordcmp = "word %s '%s'"%(cmpop, w)
            res = conn.query("""SELECT freq / (1.0*maxfreq), value, cond FROM ft_voc_%(name)s """
                             """ INNER JOIN ft_doc_%(name)s ON ft_doc_%(name)s.id = ft_voc_%(name)s.docid """
                             """ INNER JOIN ft_params_%(name)s ON ft_voc_%(name)s.word = ft_params_%(name)s.name """
                             """ WHERE %(wordcmp)s """
                             """ ORDER BY freq DESC LIMIT %(maxsize)s"""%locals())
            qtf = qfreq/qmaxfreq 
            for dtf, ni, cond in res:
                ni = float(ni)
                idf = math.log(ndocs / ni)
                dw = dtf * idf
                qw = qtf * idf
                try:
                    x = docres[cond]
                    x[0] += dw * qw
                    x[1] += dw * dw
                    x[2] += qw * qw
                    x[3] += 1
                except KeyError:
                    docres[cond] = [dw * qw, dw * dw, qw * qw, 1]
        res = []
        for cond, (sumdwdq, sumsquaredw, sumsquareqw, freq) in docres.items():
            # print (sumdwdq, sumsquaredw, sumsquareqw, freq)
            res.append((freq * sumdwdq / (1.0+(math.sqrt(sumsquaredw) * math.sqrt(sumsquareqw))), cond))
        res.sort(reverse=True)
        # print res
        return res
